# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

import sys
import time

import jax
from jax import config
import jax.numpy as jnp
import flax

from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models, assembler

config.update("jax_enable_x64", True)
config.update("jax_compilation_cache_dir", './cache')

### Dirichlet problem of transient heat conduction equation (moving least square solution structure)
# Comparison of different linear solvers


# Positive smooth distance functions that select the specific boundary segment
# Time dimension is always the last dimension (here the second)
psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([0.,0.]), jnp.stack([1.,0.]))
psdf_right = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([1.,0.]), jnp.stack([1.,1.]))
psdf_left = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([0.,1.]), jnp.stack([0.,0.]))
psdfs = [psdf_bottom, psdf_right, psdf_left]

# Positive smooth distance funcion of full domain (used for seeding nodes and integration points)
psdf_rect = lambda x: geometry.psdf_parallelogram(x, jnp.stack([0.,0.]), jnp.stack([0.,1.]), jnp.stack([1.,0.]))
psdf_domain = psdf_rect

# Initial condition
def bc_bottom(x):
  return jax.lax.sin(jnp.pi * x[0])

# Dirichlet boundary conditions on left an right side
def bc_left(x):
  return 0.
def bc_right(x):
  return 0.

# Assignment of boundary conditions (has to be in the same order as the psdfs)
dirichlet_conditions = [bc_bottom, bc_right, bc_left]

# Transfinite interpolation of boundary segments
b_cond = lambda x: geometry.transfinite_interpolation(x, psdfs, dirichlet_conditions)
psdf = lambda x: geometry.psdf_unification(x, psdfs)

### Initialization and generation of nodes and integration points
nod_spacing = 0.02
(x_nodes, n_nodes, region_size) = seeder.regular_in_psdf(psdf_domain, [0.,0.], [1.,1.], nod_spacing)

# Integration points
int_spacing = nod_spacing
(x_int, w_int, n_int) = seeder.gauss_points_in_psdf(psdf_domain, [0.,0.], [1.,1.], int_spacing, order=2)

# Support radius for moving least square ansatz
support_radius = 3.51 * nod_spacing

# Definition of pde
diffusivity = lambda x: 0.5
pde = models.heat_equation(diffusivity_fun=diffusivity)

### Neighborhood search
(num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)


### Settings
n_fields = 1
dofs0 = jnp.zeros((n_nodes, n_fields))
static_settings = flax.core.FrozenDict({
  'solution space': ('mls',),
  'shape function mode': 'compiled',
  'number of fields': (n_fields,),
  'assembling mode': ('sparse',),
  'maximal number of neighbors': (max_neighbors,),
  'order of basis functions': (2,),
  'weight function type': ('gaussian',),
  'variational scheme': ('least square pde loss',),
  'model': (pde,),
  'solution structure': ('dirichlet',),
  'boundary conditions': (b_cond,),
  'psdf': (psdf,),

  # See different solver options below.
  # Remarks:
  # For small to medium problems on cpu, pardiso is a good choice
  # If tangent can not be loaded in RAM: try matrix-free solvers
  # For very large problems: try multigrid preconditioners on (gpu) cluster (petsc, pyamg)
  # For itterative solvers: setting tol or maxiter to low can reduce the solution accuracy

  # 'solver type': 'linear',
  # 'solver backend': 'pyamg',                # for further options see pyamg documentation
  # 'solver': 'cg',                           # cg, bcgs
  # 'type of preconditioner': 'ruge stuben',  # ruge stuben, smoothed aggregation 

  # 'solver type': 'linear',
  # 'solver backend': 'petsc',        # available options see PETSc documentation
  # 'solver': 'bcgs',
  # 'type of preconditioner': 'ilu',

  # 'solver type': 'linear',
  # 'solver backend': 'jax',              # if possible, uses matrix-free algorithms
  # 'solver': 'cg',                       #cg, bicgstab, gmres(, cg normal, lu, cholesky, qr, jacobi)
  # 'type of preconditioner': 'jacobi',   # 'jacobi', 'none'
  # 'hvp type': 'fwdrev',                 # fwdrev, revrev (only for symmetric matrices), linearize, assemble

  # 'solver type': 'linear',
  # 'solver backend': 'scipy',  # only on cpu
  # 'solver': 'lapack',         # lapack, umfpack

  'solver type': 'linear',
  'solver backend': 'pardiso', # only on cpu
  'solver': 'lu', # lu, qr

  # # Only works for problems where a functional can be minimized (e.g. least square pde loss, least square function approximation)
  # 'solver type': 'minimize',  # maxiter and tol may have to be adjusted (see effect on optimized L2 error); 
  # 'solver': 'lbfgs',          # lbfgs, gradient descent, nonlinear cg, bfgs, gauss newton, levenberg marquart

  'verbose': 2,
  'connectivity': (utility.jnp_to_tuple(neighbor_list),),
})
settings = {
  'beta': (3,),
  'node coordinates': x_nodes,
  'integration coordinates': (x_int,),
  'integration weights': (w_int,),
  'support radius': (support_radius,),
}


### Precompute shape functions and pass them to e.g. static_settings...
settings = solution_structures.precompile(dofs0, settings, static_settings)


### Call solver
dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8, maxiter=1000)


### Preperation of postprocessing data
local_dofs = dofs[neighbor_list]
def post_fun(x, itt, local_dofs, settings, static_settings):
  return solution_structures.solution_structure(x, itt, local_dofs, settings, static_settings, 0)

post_fun_vj = jax.jit(jax.vmap(post_fun, (0,0,0,None,None)), static_argnames=['static_settings'])
data = post_fun_vj(x_int, jnp.arange(n_int), local_dofs, settings, static_settings)


# Plotting with pyvista
plotter.pv_plot(x_int, data[:,0], export_vtk=False, show=True)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.


"""
These functions are a modification and extension of the functions in jaxopt._src.implicit_diff for
external solve functions with a jax.experimental.sparse.BCOO matrix as an argument instead of a matvec function.
The root_vjp and root_jvp functions were modified in a way that external solvers can be used via a pure_callback.
With the wrapper custom_root, root solvers can be made differentiable both in forward or reverse mode of arbitrary order.
Mixing the differentiation mode is currently not possible.
"""

import inspect
from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple

import jax
import jax.numpy as jnp
import numpy as np

def tree_scalar_mul(scalar, tree_x):
  """Compute scalar * tree_x."""
  return jax.tree_util.tree_map(lambda x: scalar * x, tree_x)

def _extract_kwargs(kwarg_keys, flat_args):
  n = len(flat_args) - len(kwarg_keys)
  args, kwarg_vals = flat_args[:n], flat_args[n:]
  kwargs = dict(zip(kwarg_keys, kwarg_vals))
  return args, kwargs

def _signature_bind(signature, *args, **kwargs):
  ba = signature.bind(*args, **kwargs)
  ba.apply_defaults()
  return ba.args, ba.kwargs

def _signature_bind_and_match(signature, *args, **kwargs):
  # We want to bind *args and **kwargs based on the provided
  # signature, but also to associate the resulting positional
  # arguments back. To achieve this, we lift arguments to a triple:
  #
  #   (was_kwarg, ref, value)
  #
  # where ref is an index position (int) if the original argument was
  # from *args and a dictionary key if the original argument was from
  # **kwargs. After binding to the inspected signature, we use the
  # tags to associate the resolved positional arguments back to their
  # arg and kwarg source.

  args = [(False, i, v) for i, v in enumerate(args)]
  kwargs = {k: (True, k, v) for (k, v) in kwargs.items()}
  ba = signature.bind(*args, **kwargs)

  mapping = [(was_kwarg, ref) for was_kwarg, ref, _ in ba.args]

  def map_back(out_args):
    src_args = [None] * len(args)
    src_kwargs = {}
    for (was_kwarg, ref), out_arg in zip(mapping, out_args):
      if was_kwarg:
        src_kwargs[ref] = out_arg
      else:
        src_args[ref] = out_arg
    return src_args, src_kwargs

  out_args = tuple(v for _, _, v in ba.args)
  out_kwargs = {k: v for k, (_, _, v) in ba.kwargs.items()}
  return out_args, out_kwargs, map_back

def _jvp_args(residual_fun, sol, args, tangents):
  """JVP in the second argument of residual_fun."""
  # We close over the solution.
  fun = lambda *y: residual_fun(sol, *y)
  return jax.jvp(fun, args, tangents)[1]

def _root_vjp(residual_fun: Callable,
             mat_fun: Callable,
             sol: Any,
             args: Tuple,
             cotangent: Any,
             solve_fun: Callable,
             free_dofs: Any,
             ) -> Any:
  """Vector-Jacobian product of a root.

  The invariant is ``residual_fun(sol, *args) == 0``.

  Args:
    residual_fun: the optimality function to use.
    sol: solution / root (pytree).
    mat_fun: a function that has to compute the sparse tangent matrix with sol and args as arguments.
    args: tuple containing the arguments with respect to which we wish to
      differentiate ``sol`` against.
    cotangent: vector to left-multiply the Jacobian with
      (pytree, same structure as ``sol``).
    solve_fun: a linear solver of the form ``x = solve_fun(mat, b)``,
      where ``mat`` is as jax.experimental.sparse.BCOO matrix.
  Returns:
    tuple of the same length as ``len(args)`` containing the vjps w.r.t.
    each argument. Each ``vjps[i]`` has the same pytree structure as
    ``args[i]``.
  """

  @jax.custom_vjp
  def linear_solver_fun_vjp(A, b):
      return solve_fun(A, b) # Here can be an external callback

  def linear_solver_fun_vjp_fwd(A, b):
      result = linear_solver_fun_vjp(A, b)
      return result, (A, b, result)

  def linear_solver_fun_vjp_bwd(res, g):
      A, b, result = res

      # Ensure g matches the size of the expected gradient
      if free_dofs is not None:

          # Sparse outer product
          result_dot = linear_solver_fun_vjp(A.T, g)

          size = A.shape[0]
          idx = jnp.arange(size)[free_dofs]
          result_tmp = jnp.zeros((size,), dtype=jnp.float_)
          result_tmp = result_tmp.at[idx].set(result)
          result_dot_tmp = jnp.zeros((size,), dtype=jnp.float_)
          result_dot_tmp = result_dot_tmp.at[idx].set(result_dot)

          # # Use the sparsity structure of A to build Fx_dot
          indices = jnp.asarray(A.indices, dtype=jnp.int64)
          data = -result_dot_tmp[indices[:, 0]] * result_tmp[indices[:, 1]]

          # Construct Fx_dot with the same sparsity pattern
          Fx_dot = jax.experimental.sparse.BCOO((data, indices), shape=A.shape)
          Fy_dot = result_dot
      else:
          # Sparse outer product
          result_dot = linear_solver_fun_vjp(A.T, g)

          # Use the sparsity structure of A to build Fx_dot
          indices = jnp.asarray(A.indices, dtype=jnp.int64)
          data = -result_dot[indices[:, 0]] * result[indices[:, 1]]

          # Construct Fx_dot with the same sparsity pattern
          Fx_dot = jax.experimental.sparse.BCOO((data, indices), shape=A.shape)
          Fy_dot = result_dot

      # Correctly return the shapes that match the input arguments
      return (Fx_dot, Fy_dot)

  linear_solver_fun_vjp.defvjp(linear_solver_fun_vjp_fwd, linear_solver_fun_vjp_bwd)
  diffable_solve_fun = linear_solver_fun_vjp

  mat = mat_fun(sol, *args)
  
  # The solution of A^T u = v, where
  # A = jacobian(residual_fun, argnums=0)
  # v = -cotangent.
  v = tree_scalar_mul(-1, cotangent)

  if free_dofs is not None:
    constrained_dofs = np.invert(free_dofs)
    idx_imposed = jnp.arange(sol.flatten().shape[0])[constrained_dofs]

    v_flat = v.flatten()[free_dofs]
    u = jnp.zeros_like(v)
    u_flat = u.flatten()
    idx_free = jnp.arange(u_flat.shape[0])[free_dofs]
    u_f = diffable_solve_fun(mat.T, v_flat)
    u_flat = u_flat.at[idx_free].set(u_f)
   
    def fun_args(*args):
      def residual_fun_tmp(sol, *args):
        dirichlet_values = args[0]['dirichlet conditions'].flatten()[constrained_dofs]
        sol_with_bc = sol.flatten()
        sol_with_bc = sol_with_bc.at[idx_imposed].set(dirichlet_values)
        return residual_fun(sol_with_bc.reshape(sol.shape), *args)
      return residual_fun_tmp(sol, *args) 

    u = u_flat.reshape(v.shape)
  else:
    def fun_args(*args):
      return residual_fun(sol, *args)

    v_flat = v.flatten()
    u = diffable_solve_fun(mat.T, v_flat).reshape(v.shape)

  _, vjp_fun_args = jax.vjp(fun_args, *args)
  args_vjp = vjp_fun_args(u)

  if free_dofs is not None:
    dirichlet_cotangent = cotangent.flatten()[constrained_dofs]
    updated_args0 = args_vjp[0]
    tmp = updated_args0['dirichlet conditions'].flatten()
    tmp = tmp.at[constrained_dofs].add(dirichlet_cotangent)
    updated_args0['dirichlet conditions'] = tmp.reshape(updated_args0['dirichlet conditions'].shape)
    args_vjp = (updated_args0,) + args_vjp[1:]

  return args_vjp

def _root_jvp(residual_fun: Callable,
             mat_fun: Callable,
             sol: Any,
             args: Tuple,
             tangents: Tuple,
             solve_fun: Callable,
             free_dofs: Any
             ) -> Any:
  """
  Jacobian-vector product of a root.

  The invariant is ``residual_fun(sol, *args) == 0``.

  Args:
    residual_fun: the optimality function to use.
    mat_fun: a function that has to compute the sparse tangent matrix with sol and args as arguments.
    sol: solution / root (pytree).
    args: tuple containing the arguments with respect to which to differentiate.
    tangents: a tuple of the same size as ``len(args)``. Each ``tangents[i]``
      has the same pytree structure as ``args[i]``.
    solve_fun: a linear solver of the form ``x = solve_fun(mat, b)``,
      where ``mat`` is as jax.experimental.sparse.BCOO matrix.
  Returns:
    a pytree with the same structure as ``sol``.
  """
  # Compute tangent matrix
  A = mat_fun(sol, *args)
  mat_shape = A.shape

  # Forward differentiable sparse linear solver
  @jax.custom_jvp
  def linear_solver_fun_jvp(data, indices, b):
    A = jax.experimental.sparse.BCOO((data, indices), shape=mat_shape)
    return solve_fun(A, b)
    
  @linear_solver_fun_jvp.defjvp
  def linear_solver_fun_jvp_rule(primals, tangents):
    data, indices, b = primals
    data_dot, _, b_dot = tangents

    # Compute the primal result using the linear solver function
    primal_result = linear_solver_fun_jvp(data, indices, b)

    A_dot = jax.experimental.sparse.BCOO((data_dot, indices), shape=mat_shape)

    # Handle the tangent calculation
    if free_dofs is not None:
      # Calculate the tangent result
      primal_result_tmp = jnp.zeros((mat_shape[0],), dtype=jnp.float64)
      idx = jnp.arange(mat_shape[0])[free_dofs]
      primal_result_tmp = primal_result_tmp.at[idx].set(primal_result)
      rhs = b_dot - (A_dot @ primal_result_tmp)[free_dofs]
      result_dot = linear_solver_fun_jvp(data, indices, rhs)
    else:
      result_dot = linear_solver_fun_jvp(data, indices, b_dot - A_dot @ primal_result)

    return primal_result, result_dot

  # Assign the jvp-enabled solver function
  solve_func = linear_solver_fun_jvp

  if free_dofs is not None:
    constrained_dofs = np.invert(free_dofs)
    idx_imposed = jnp.arange(A.shape[0])[constrained_dofs]
    idx_free = jnp.arange(A.shape[0])[free_dofs]

    # Explicit imposition of DOFs in order to be able to take derivatives w.r.t. nodally imposed DOFs
    def residual_fun_tmp(sol, *args):
      dirichlet_values = args[0]['dirichlet conditions'].flatten()[constrained_dofs]
      sol_with_bc = sol.flatten()
      sol_with_bc = sol_with_bc.at[idx_imposed].set(dirichlet_values)
      return residual_fun(sol_with_bc.reshape(sol.shape), *args)

    Bv = _jvp_args(residual_fun_tmp, sol, args, tangents)
    Bv_free = Bv.flatten()[free_dofs]
    Jv_free = solve_func(A.data, A.indices, -Bv_free)
    Jv_flat = jnp.zeros_like(sol).flatten()
    Jv_flat = Jv_flat.at[idx_free].set(Jv_free)

    dirichlet_values_dot = tangents[0]['dirichlet conditions'].flatten()[constrained_dofs]
    Jv_flat = Jv_flat.at[idx_imposed].set(dirichlet_values_dot)

  else:
    Jv_flat = solve_func(A.data, A.indices, -Bv.flatten())

  # Repack Jv_free into the original structure of tangents
  Jv = Jv_flat.reshape(Bv.shape)
  return Jv

def _custom_root(solver_fun, 
                 residual_fun, 
                 mat_fun, 
                 free_dofs, 
                 solve, 
                 has_aux,
                 mode='reverse', 
                 reference_signature=None):
  # When caling through `jax.custom_vjp`, jax attempts to resolve all
  # arguments passed by keyword to positions (this is in order to
  # match against a `nondiff_argnums` parameter that we do not use
  # here). It does so by resolving them according to the custom_jvp'ed
  # function's signature. It disallows functions defined with a
  # catch-all `**kwargs` expression, since their signature cannot
  # always resolve all keyword arguments to positions.
  #
  # We can loosen the constraint on the signature of `solver_fun` so
  # long as we resolve keywords to positions ourselves. We can do so
  # just in time, by flattening the `kwargs` dict (respecting its
  # iteration order) and supplying `custom_vjp` with a
  # positional-argument-only function. We then explicitly coordinate
  # flattening and unflattening around the `custom_vjp` boundary.
  #
  # Once we make it past the `custom_vjp` boundary, we do some more
  # work to align arguments with the reference signature (which is, by
  # default, the signature of `residual_fun`).

  solver_fun_signature = inspect.signature(solver_fun)

  if reference_signature is None:
    reference_signature = inspect.signature(residual_fun)

  elif not isinstance(reference_signature, inspect.Signature):
    # If is a CompositeLinearFunction, accesses subfun.
    # Otherwise, assumes a Callable.
    fun = getattr(reference_signature, "subfun", reference_signature)
    reference_signature = inspect.signature(fun)

  def make_custom_solver_fun(solver_fun, kwarg_keys):

    def solver_fun_flat_tmp(*flat_args):
      args, kwargs = _extract_kwargs(kwarg_keys, flat_args)
      return solver_fun(*args, **kwargs)

    if mode == 'reverse' or mode == 'backward':
      solver_fun_flat = jax.custom_vjp(solver_fun_flat_tmp)
    elif mode == 'forward':
      solver_fun_flat = jax.custom_jvp(solver_fun_flat_tmp)
    else:
      raise ValueError("Mode must be either 'forward' or 'reverse'.")

    # Forward-mode differentiation (JVP)
    def solver_fun_jvp(primals, tangents):
      args, kwargs = _extract_kwargs(kwarg_keys, primals)
      tangent_args, tangent_kwargs = _extract_kwargs(kwarg_keys, tangents)

      # Compute the primal solution using the root solver function
      primal_sol = solver_fun_flat(*args, **kwargs)

      # Handle has_aux case
      if has_aux:
        sol = primal_sol[0]
        aux_data = primal_sol[1:]
      else:
        sol = primal_sol

      # Compute JVP using root_jvp
      jvp_sol = _root_jvp(
        residual_fun=residual_fun,
        mat_fun=mat_fun,
        sol=sol,
        args=args[1:],  # Exclude the initial params from args
        tangents=tangent_args[1:],  # Exclude the initial params from tangents
        solve_fun=solve,
        free_dofs=free_dofs)

      if has_aux:
        # Return primal and tangent for both solution and auxiliary data
        aux_tangent = jax.tree_util.tree_map(jnp.zeros_like, aux_data)
        return (sol,) + aux_data, (jvp_sol,) + aux_tangent
      else:
        return primal_sol, jvp_sol

    # Reverse-mode differentiation (VJP)
    def solver_fun_fwd(*flat_args):
      res = solver_fun_flat(*flat_args)
      return res, (res, flat_args)

    def solver_fun_bwd(tup, cotangent):
      res, flat_args = tup
      args, kwargs = _extract_kwargs(kwarg_keys, flat_args)

      # solver_fun can return auxiliary data if has_aux = True.
      if has_aux:
        cotangent = cotangent[0]
        sol = res[0]
      else:
        sol = res

      ba_args, ba_kwargs, map_back = _signature_bind_and_match(
          reference_signature, *args, **kwargs)
      if ba_kwargs:
        raise TypeError(
          "keyword arguments to solver_fun could not be resolved to "
          "positional arguments based on the signature "
          f"{reference_signature}. This can happen under custom_root if "
          "residual_fun takes catch-all **kwargs, or under "
          "custom_fixed_point if fixed_point_fun takes catch-all **kwargs, "
          "both of which are currently unsupported.")

      # Compute VJPs w.r.t. args.
      vjps = _root_vjp(residual_fun=residual_fun, 
                      mat_fun=mat_fun, 
                      sol=sol,
                      args=ba_args[1:], 
                      cotangent=cotangent, 
                      solve_fun=solve, 
                      free_dofs=free_dofs)
      # Prepend None as the vjp for init_params.
      vjps = (None,) + vjps

      arg_vjps, kws_vjps = map_back(vjps)
      ordered_vjps = tuple(arg_vjps) + tuple(kws_vjps[k] for k in kwargs.keys())
      return ordered_vjps

    if mode == 'reverse' or mode == 'backward':
      solver_fun_flat.defvjp(solver_fun_fwd, solver_fun_bwd)
    else:
      solver_fun_flat.defjvp(solver_fun_jvp)
    return solver_fun_flat

  def wrapped_solver_fun(*args, **kwargs):
    args, kwargs = _signature_bind(solver_fun_signature, *args, **kwargs)
    keys, vals = list(kwargs.keys()), list(kwargs.values())
    return make_custom_solver_fun(solver_fun, keys)(*args, *vals)

  return wrapped_solver_fun

def custom_root(residual_fun: Callable,
                mat_fun: Callable,
                solve: Callable,
                free_dofs: Any = None,
                has_aux: bool = False,
                mode='reverse',
                reference_signature: Optional[Callable] = None):
  """Decorator for adding implicit differentiation to a root solver.

  Args:
    residual_fun: A callable the returns the possibly nonlinear residual of which to find the root of, 
      ``residual_fun(dofs, *args)``.
      The invariant is ``residual_fun(sol, *args) == 0`` at the solution / root ``sol``.
    mat_fun: A callable that returns the sparse tangent matrix as a jax.experimental.BCOO with dofs and 
      args as arguments. Can also be a pure callback.
    solve: A linear solver of the form ``solve(mat[jax.experimental.BCOO], b[jnp.ndarray])``.
    has_aux: whether the decorated root solver function returns auxiliary data.
    mode: The differentiation mode ('forward' or 'reverse'/'backward').
    reference_signature: optional function signature
      (i.e. arguments and keyword arguments), with which the
      solver and optimality functions are expected to agree. Defaults
      to ``residual_fun``. It is required that solver and optimality
      functions share the same input signature, but both might be
      defined in such a way that the signature correspondence is
      ambiguous (e.g. if both accept catch-all ``**kwargs``). To
      satisfy custom_root's requirement, any function with an
      unambiguous signature can be provided here.

  Returns:
    The decorated root solver function that is equipped a with custom vjp or jvp rule.

  Example:
    See e.g. the implementation of autopdex.solver.adaptive_load_stepping.
  """

  def wrapper(solver_fun):
    return _custom_root(solver_fun, residual_fun, mat_fun, free_dofs, solve, has_aux, mode,
                        reference_signature)

  return wrapper



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

def test_example_poisson():
    from jax import config
    import jax.numpy as jnp
    import flax

    from autopdex import seeder, geometry, solution_structures, utility, models, assembler

    config.update("jax_enable_x64", True)

    # Dirichlet problem of the Poisson equation in 2d: [0,1]x[0,1]
    # Demonstration of compile time reduction by manually pre-compiling solution structure


    # Positive smooth distance functions that select the specific boundary segments
    psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, [0.,0.], [1.,0.])
    psdf_right = lambda x: geometry.psdf_trimmed_line(x, [1.,0.], [1.,1.])
    psdf_top = lambda x: geometry.psdf_trimmed_line(x, [1.,1.], [0.,1.])
    psdf_left = lambda x: geometry.psdf_trimmed_line(x, [0.,1.], [0.,0.])
    psdf_rect = lambda x: geometry.psdf_parallelogram(x, [0.,0.], [0.,1.], [1.,0.])
    psdfs = [psdf_bottom, psdf_right, psdf_top, psdf_left]

    # Positive smooth distance funcion of full domain (used for seeding nodes and integration points)
    psdf_domain = psdf_rect

    # Boundary condition functions
    def homogeneous(x): # Bottom, right and top boundaries
        return 0

    # Assignment of boundary conditions
    dirichlet_conditions = [homogeneous, homogeneous, homogeneous, homogeneous]


    ### Transfinite interpolation of boundary segments
    b_cond = lambda x: geometry.transfinite_interpolation(x, psdfs, dirichlet_conditions)
    psdf = psdf_domain

    ### Initialization and generation of nodes and integration points
    # Regular node placement
    nod_spacing = 0.2
    (x_nodes, n_nodes, region_size) = seeder.regular_in_psdf(psdf_domain, [0.,0.], [1.,1.], nod_spacing)

    # Integration points
    # Collocation, Note: this can in some cases cause bad conditioning, unphysical low-energy modes and divergence. 
    # If so, use more integration points, e.g. Gauss integration on a background grid (seeder.gauss_points_in_psdf)
    x_int = x_nodes
    n_int = n_nodes

    # Support radius for moving least square ansatz
    support_radius = 2.01 * nod_spacing

    # Integration point weights
    w_int = (region_size / n_int) * jnp.ones(n_int)

    # Definition of pde
    pde = models.poisson(source_fun = lambda x: 1)


    ### Neighborhood search
    (num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)


    ### Settings
    n_fields = 1
    dofs0 = jnp.zeros((n_nodes, n_fields))
    static_settings = flax.core.FrozenDict({

    # Precompilation of the solution structure at the integration points can considerable improve compile time
    'shape function mode': 'direct',
    # 'shape function mode': 'compiled', # precompilation is done in line 107
    
    'solution space': ('mls',),
    'number of fields': (n_fields,),
    'assembling mode': ('sparse',),
    'maximal number of neighbors': (max_neighbors,),
    'order of basis functions': (1,),
    'weight function type': ('bump',),
    'variational scheme': ('strong form galerkin',),
    'model': (pde,),
    'solution structure': ('dirichlet',),
    'boundary conditions': (b_cond,),
    'psdf': (psdf,),
    'solver type': 'linear',

    'solver backend': 'scipy',
    'solver': 'lapack',
    # 'solver backend': 'jax',
    # 'solver': 'bicgstab',
    # 'type of preconditioner': 'jacobi',
    # 'hvp type': 'fwdrev',

    'verbose': 2,
    'connectivity': (utility.jnp_to_tuple(neighbor_list),),
    })
    settings = {
    'node coordinates': x_nodes,
    'beta': (3,),
    'integration coordinates': (x_int,),
    'integration weights': (w_int,),
    'support radius': (support_radius,),
    }

    tangent = assembler.assemble_tangent(dofs0, settings, static_settings)
    check = tangent.data.flatten().sum()
    # print(check)
    assert jnp.isclose(check, -0.9242697917022855)



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.


def test_least_square_fem():
    from jax import config
    import jax.numpy as jnp
    import flax
    import pygmsh

    from autopdex import seeder, geometry, utility, models, assembler

    config.update("jax_enable_x64", True)

    # Cook's membrane problem: first order system least square finite elements (T2/Tri-6) as Rvachev function solution structure



    ### Definition of geometry and boundary conditions
    # Positive smooth distance functions that select the specific boundary segments
    pts = [[0.,0.], [48.,44.], [48.,60.], [0.,44.]]
    psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, pts[0], pts[1])
    psdf_right = lambda x: geometry.psdf_trimmed_line(x, pts[1], pts[2])
    psdf_top = lambda x: geometry.psdf_trimmed_line(x, pts[2], pts[3])
    psdf_left = lambda x: geometry.psdf_trimmed_line(x, pts[3], pts[0])
    psdfs = [psdf_bottom, psdf_right, psdf_top, psdf_left]

    # Positive smooth distance funcion of full domain
    sdf_cooks = lambda x: geometry.sdf_convex_polygon_2d(x, pts)
    psdf_cooks = lambda x: geometry.only_positive(sdf_cooks(x))
    psdf = lambda x: jnp.asarray([psdf_cooks(x), psdf_cooks(x)])

    # Boundary condition functions
    homogeneous = lambda x: 0.
    normals = lambda x: geometry.normals_from_normalized_sdf(x, psdf_cooks)

    # Field selectors for boundary conditions
    selection_dirichlet_1 = lambda x: jnp.asarray([1.,0.,0.,0.,0.,0.])
    selection_dirichlet_2 = lambda x: jnp.asarray([0.,1.,0.,0.,0.,0.])
    def selection_neumann_1(x):
        normal_x = normals(x)[0]
        normal_y = normals(x)[1]
        return jnp.asarray([0.,0.,normal_x,normal_y,0.,0.])
    def selection_neumann_2(x):
        normal_x = normals(x)[0]
        normal_y = normals(x)[1]
        return jnp.asarray([0.,0.,0.,0.,normal_x,normal_y])

    # Assignment of boundary conditions (same ordering as in psdfs)
    load = lambda x: 4.0
    bc_1 = [homogeneous, homogeneous, homogeneous, homogeneous]
    bc_2 = [homogeneous, load, homogeneous, homogeneous]
    a_1 = [selection_neumann_1, selection_neumann_1, selection_neumann_1, selection_dirichlet_1]
    a_2 = [selection_neumann_2, selection_neumann_2, selection_neumann_2, selection_dirichlet_2]

    ### Transfinite interpolation of boundary segments
    b_coeff = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, a_1),
                                    geometry.transfinite_interpolation(x, psdfs, a_2)])
    b_cond = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, bc_1),
                                    geometry.transfinite_interpolation(x, psdfs, bc_2)])

    ### Discretization
    ### Definition of geometry and boundary conditions
    with pygmsh.geo.Geometry() as geom:
        elem_length = 50.
        geom.add_polygon(pts,mesh_size=elem_length)
        mesh = geom.generate_mesh(order=2)

    # Import region mesh
    n_dim = 2
    x_nodes = jnp.asarray(mesh.points[:,:n_dim])
    n_nodes = x_nodes.shape[0]
    elements = mesh.cells_dict['triangle6']

    # Generate integration points in mesh
    (x_int, w_int, n_int, domain_connectivity) = seeder.int_pts_in_tri_mesh(x_nodes, elements, order=3)

    # Definition of pde
    lam = 100.
    mu = 40.
    Em = mu * (3 * lam + 2 * mu) / (lam + mu)
    nu = lam / (2 * (lam + mu))
    youngs_mod_fun = lambda x: Em
    poisson_ratio_fun = lambda x: nu
    # pde = models.linear_elasticity_fos(youngs_mod_fun = youngs_mod_fun, poisson_ratio_fun = poisson_ratio_fun, mode='plain strain', spacing=elem_length)
    pde = models.hyperelastic_steady_state_fos(models.neo_hooke, youngs_mod_fun = youngs_mod_fun, poisson_ratio_fun = poisson_ratio_fun, mode='plain strain', spacing=elem_length)

    ### Setting
    n_fields = 6
    dofs0 = jnp.zeros((n_nodes, n_fields))
    static_settings = flax.core.FrozenDict({
    'solution space': ('fem simplex',),
    'shape function mode': 'direct',
    'number of fields': (n_fields,),
    'assembling mode': ('sparse',),
    'maximal number of neighbors': (domain_connectivity.shape[-1],),
    'variational scheme': ('least square pde loss',),
    'model': (pde,),
    'solution structure': ('first order set',),
    'boundary coefficients': (b_coeff,),
    'boundary conditions': (b_cond,),
    'psdf': (psdf,),
    'solver type': 'newton',
    'solver backend': 'scipy',
    'solver': 'lapack',
    'verbose': 0,
    'connectivity': (utility.jnp_to_tuple(domain_connectivity),),
    })
    settings = {
    'node coordinates': x_nodes,
    'integration coordinates': (x_int,),
    'integration weights': (w_int,),
    }

    check = assembler.assemble_tangent(dofs0, settings, static_settings).data.flatten().sum()
    # print(check)
    assert jnp.isclose(check, 1387.082244521068)




# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
In this example, the wave equation is solved on the square domain 
:math:`(x, t)` with :math:`x \in (0, 1)` and :math:`t \in (0, 1)`.
The problem is formulated as a first-order system, and the solution 
is searched using a least-squares variational method. For the solution space, 
meshfree moving least squares (MLS) shape functions of second order are used, 
with integration performed on a background grid (as in the Element-Free 
Galerkin (EFG) method) using Gaussian quadrature. The application of boundary
conditions is achieved through the solution structure for first-order
systems using Rvachev distance functions to the respective boundary segments.
"""

if __name__ == "__main__":
  import sys

  import jax
  from jax import config
  import jax.numpy as jnp
  import flax

  from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models

  config.update("jax_enable_x64", True)
  config.update("jax_compilation_cache_dir", './cache')



  ### Definition of geometry and boundary conditions
  # Positive smooth distance functions that select the specific boundary segment
  # Time dimension is always the last dimension (here the second)
  psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([0.,0.]), jnp.stack([1.,0.]))
  psdf_right = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([1.,0.]), jnp.stack([1.,1.]))
  psdf_left = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([0.,1.]), jnp.stack([0.,0.]))
  psdfs_1 = [psdf_bottom, psdf_right, psdf_left]
  psdfs_2 = [psdf_bottom, psdf_right, psdf_left]


  # Initial condition
  def bc_bottom1(x): # Initial value
    return jax.lax.exp(-100. * (x[0]-0.3)**2)
  def bc_bottom2(x): # Initial speed
    return 0.

  # Dirichlet boundary conditions on left an right side
  def bc_left(x):
    return 0.
  def bc_right(x):
    return 0.
  bcs_1 = [bc_bottom1, bc_right, bc_left]
  bcs_2 = [bc_bottom2]

  # Field selector for boundary conditions
  def dirichlet(x):
    return jnp.asarray([1., 0., 0.])
  def neumann_x(x):
    return jnp.asarray([0., 1., 0.])
  def neumann_t(x):
    return jnp.asarray([0., 0., 1.])
  ai_1 = [dirichlet, dirichlet, neumann_x]
  ai_2 = [neumann_t, neumann_t, neumann_t] # Ensure orthogonality of the coefficients along the boundary segments!

  # Transfinite interpolation of boundary segments
  b_coeff = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs_1, ai_1),
                                    geometry.transfinite_interpolation(x, psdfs_2, ai_2)])
  b_cond = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs_1, bcs_1),
                                  bcs_2[0](x)])
  psdf = lambda x: jnp.asarray([geometry.psdf_unification(x, psdfs_1),
                                psdf_bottom(x)]) # Only select the region where boundary conditions shall be enforced!

  # Positive smooth distance funcion of full domain (used for seeding nodes and integration points)
  psdf_rect = lambda x: geometry.psdf_parallelogram(x, jnp.stack([0.,0.]), jnp.stack([0.,1.]), jnp.stack([1.,0.]))
  psdf_domain = psdf_rect

  ### Initialization and generation of nodes and integration points
  nod_spacing = 0.02
  # nod_spacing = 0.01
  (x_nodes, n_nodes, region_size) = seeder.regular_in_psdf(psdf_domain, [0.,0.], [1.,1.], nod_spacing)

  # Gauss integration on background grid
  int_spacing = nod_spacing
  (x_int, w_int, n_int) = seeder.gauss_points_in_psdf(psdf_domain, [0.,0.], [1.,1.], int_spacing, order=2)

  # Support radius for moving least square ansatz
  support_radius = 3.01 * nod_spacing

  # ### Visualizer for boundary functions
  # fun = lambda x: psdf(x)[0]
  # plotter.pv_function_plot(x_int, fun)
  # sys.exit()


  # Definition of pde
  wave_number = lambda x: 1.0
  pde = models.d_alembert_fos(wave_number_fun=wave_number, spacing=nod_spacing)

  ### Neighborhood search
  (num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)


  ### Settings
  n_fields = 3
  dofs0 = jnp.zeros((n_nodes, n_fields))
  static_settings = flax.core.FrozenDict({
    'solution space': ('mls',),
    'shape function mode': 'compiled',
    'number of fields': (n_fields,),
    'assembling mode': ('sparse',),
    'maximal number of neighbors': (max_neighbors,),
    'order of basis functions': (2,),
    'weight function type': ('gaussian',),
    'variational scheme': ('least square pde loss',),
    'model': (pde,),
    'solution structure': ('first order set',),
    'boundary coefficients': (b_coeff,),
    'boundary conditions': (b_cond,),
    'psdf': (psdf,),
    'solver type': 'linear',
    'solver backend': 'pardiso',
    'solver': 'lu',
    'verbose': 1,
    'connectivity': (utility.jnp_to_tuple(neighbor_list),),
  })
  settings = {
    'beta': (3,),
    'node coordinates': x_nodes,
    'integration coordinates': (x_int,),
    'integration weights': (w_int,),
    'support radius': (support_radius,),
  }


  # Precompute shape functions
  settings = solution_structures.precompile(dofs0, settings, static_settings)


  ### Call solver
  dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8)


  ### Preperation of postprocessing data
  local_dofs = dofs[neighbor_list]
  post_fun_vj = jax.jit(jax.vmap(solution_structures.solution_structure, (0,0,0,None,None,None)), static_argnames=['static_settings', 'set'])
  data = post_fun_vj(x_int, jnp.arange(n_int), local_dofs, settings, static_settings, 0)

  # Plotting with pyvista
  plotter.pv_plot(x_int, data[:,0], export_vtk=False, show=True)
  plotter.pv_plot(x_int, data[:,1], export_vtk=False, show=True)
  plotter.pv_plot(x_int, data[:,2], export_vtk=False, show=True)

  # Export to Paraview
  import meshio
  import numpy as np
  points = x_int
  cells = [("vertex", np.array([[i] for i in range(len(points))]))]
  mesh = meshio.Mesh(
      points,
      cells,
      point_data={
          "field": data[:,0],
          "gradient": data[:,1:],
      },
  )
  mesh.write("wave_equation.vtk")



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

import time
import sys

import jax
from jax import lax, config
import jax.numpy as jnp
import flax
import pygmsh

from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models, spaces, assembler

config.update("jax_enable_x64", True)
config.update("jax_compilation_cache_dir", './cache')
config.update("jax_debug_nans", False)


# Cook's membrane problem: first order system least square finite elements (T2/Tri-6) as Rvachev function solution structure



### Definition of geometry and boundary conditions
# Positive smooth distance functions that select the specific boundary segments
pts = [[0.,0.], [48.,44.], [48.,60.], [0.,44.]]
psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, pts[0], pts[1])
psdf_right = lambda x: geometry.psdf_trimmed_line(x, pts[1], pts[2])
psdf_top = lambda x: geometry.psdf_trimmed_line(x, pts[2], pts[3])
psdf_left = lambda x: geometry.psdf_trimmed_line(x, pts[3], pts[0])
psdfs = [psdf_bottom, psdf_right, psdf_top, psdf_left]

# Positive smooth distance funcion of full domain
sdf_cooks = lambda x: geometry.sdf_convex_polygon_2d(x, pts)
psdf_cooks = lambda x: geometry.only_positive(sdf_cooks(x))
psdf = lambda x: jnp.asarray([psdf_cooks(x), psdf_cooks(x)])

# Boundary condition functions
homogeneous = lambda x: 0.
normals = lambda x: geometry.normals_from_normalized_sdf(x, psdf_cooks)

# Field selectors for boundary conditions
selection_dirichlet_1 = lambda x: jnp.asarray([1.,0.,0.,0.,0.,0.])
selection_dirichlet_2 = lambda x: jnp.asarray([0.,1.,0.,0.,0.,0.])
def selection_neumann_1(x):
  normal_x = normals(x)[0]
  normal_y = normals(x)[1]
  return jnp.asarray([0.,0.,normal_x,normal_y,0.,0.])
def selection_neumann_2(x):
  normal_x = normals(x)[0]
  normal_y = normals(x)[1]
  return jnp.asarray([0.,0.,0.,0.,normal_x,normal_y])

# Assignment of boundary conditions (same ordering as in psdfs)
load = lambda x: 4.0
bc_1 = [homogeneous, homogeneous, homogeneous, homogeneous]
bc_2 = [homogeneous, load, homogeneous, homogeneous]
a_1 = [selection_neumann_1, selection_neumann_1, selection_neumann_1, selection_dirichlet_1]
a_2 = [selection_neumann_2, selection_neumann_2, selection_neumann_2, selection_dirichlet_2]

### Transfinite interpolation of boundary segments
b_coeff = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, a_1),
                                   geometry.transfinite_interpolation(x, psdfs, a_2)])
b_cond = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, bc_1),
                                geometry.transfinite_interpolation(x, psdfs, bc_2)])

### Discretization
### Definition of geometry and boundary conditions
with pygmsh.geo.Geometry() as geom:
    elem_length = 1.
    geom.add_polygon(pts,mesh_size=elem_length)
    mesh = geom.generate_mesh(order=2)

# Mesh debugging
# mesh.write("mesh.vtk")

# Import region mesh
n_dim = 2
x_nodes = jnp.asarray(mesh.points[:,:n_dim])
n_nodes = x_nodes.shape[0]
elements = mesh.cells_dict['triangle6']

# Generate integration points in mesh
(x_int, w_int, n_int, domain_connectivity) = seeder.int_pts_in_tri_mesh(x_nodes, elements, order=3)

# # ### Visualizer for boundary functions
# fun = lambda x: b_cond(x)[1]
# plotter.pv_function_plot(x_int, fun)
# sys.exit()

# Definition of pde
lam = 100.
mu = 40.
Em = mu * (3 * lam + 2 * mu) / (lam + mu)
nu = lam / (2 * (lam + mu))
youngs_mod_fun = lambda x: Em
poisson_ratio_fun = lambda x: nu
# pde = models.linear_elasticity_fos(youngs_mod_fun = youngs_mod_fun, poisson_ratio_fun = poisson_ratio_fun, mode='plain strain', spacing=elem_length)
pde = models.hyperelastic_steady_state_fos(models.neo_hooke, youngs_mod_fun = youngs_mod_fun, poisson_ratio_fun = poisson_ratio_fun, mode='plain strain', spacing=elem_length)

### Setting
n_fields = 6
dofs0 = jnp.zeros((n_nodes, n_fields))
static_settings = flax.core.FrozenDict({
  'solution space': ('fem simplex',),
  'shape function mode': 'compiled',
  'number of fields': (n_fields,),
  'assembling mode': ('sparse',),
  'maximal number of neighbors': (domain_connectivity.shape[-1],),
  'variational scheme': ('least square pde loss',),
  'model': (pde,),
  'solution structure': ('first order set',),
  'boundary coefficients': (b_coeff,),
  'boundary conditions': (b_cond,),
  'psdf': (psdf,),
  'solver type': 'newton',
  'solver backend': 'pardiso',
  'solver': 'lu',
  'verbose': 0,
  'connectivity': (utility.jnp_to_tuple(domain_connectivity),),
})
settings = {
  'node coordinates': x_nodes,
  'integration coordinates': (x_int,),
  'integration weights': (w_int,),
}


# Precompute shape functions
settings = solution_structures.precompile(dofs0, settings, static_settings)


### Load stepping in order to provide better initial guesses for the newton method
start = time.time()
dofs = dofs0
multiplier = 0.0
n_step = 5
for i in range(n_step):
  # Update boundary conditions
  print('Start load step ', i+1)
  multiplier = multiplier + 1 / n_step
  load_mult = lambda x: multiplier * load(x)
  bc_2 = [homogeneous, load_mult, homogeneous, homogeneous]
  b_cond = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, bc_1),
                                  geometry.transfinite_interpolation(x, psdfs, bc_2)])
  settings['compiled bc'] = (solution_structures.precompute_coupled_boundary(b_cond, b_coeff, x_int, num_diff=1),)
  dofs, _ = solver.solver(dofs, settings, static_settings)



### Preperation of postprocessing data
local_dofs = dofs[domain_connectivity]

def post_fun(x, itt, local_dofs, settings, static_settings):
  return solution_structures.solution_structure(x, itt, local_dofs, settings, static_settings, 0)

post_fun_vj = jax.jit(jax.vmap(post_fun, (0,0,0,None,None)), static_argnames=['static_settings'])
data = post_fun_vj(x_int, jnp.arange(n_int), local_dofs, settings, static_settings)

x_vis = x_int + jnp.transpose(jnp.asarray([data[:,0],data[:,1]]))

# Plotting with pyvista
plotter.pv_plot(x_vis, data[:,0], export_vtk=False, show=True, data_on_z=False)
plotter.pv_plot(x_vis, data[:,1], export_vtk=False, show=True, data_on_z=False)
plotter.pv_plot(x_vis, data[:,2], export_vtk=False, show=True, data_on_z=False)
plotter.pv_plot(x_vis, data[:,3], export_vtk=False, show=True, data_on_z=False)
plotter.pv_plot(x_vis, data[:,4], export_vtk=False, show=True, data_on_z=False)
plotter.pv_plot(x_vis, data[:,5], export_vtk=False, show=True, data_on_z=False)



# models.py
# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
This module contains premade jax-transformable partial differential equation (PDE) formulations, weak forms, user-defined elements, and time integration procedures.

The models-functions return callables with appropriate arguments for the use required in the other modules of AutoPDEx. These functions can then be passed e.g. to the solver via static_settings['model'].

For instance, a typical weak form function has the following arguments:
  - x (jnp.ndarray): Spatial coordinates at the integration point.
  - ansatz (function): Ansatz function representing the field variable (e.g., displacement or temperature).
  - test_ansatz (function): Test ansatz function representing the virtual displacement or virtual temperature.
  - settings (dict): Settings for the computation.
  - static_settings (dict): Static settings for the computation.
  - int_point_number (int): Integration point number.
  - set: Number of domain.
"""

from inspect import signature

import jax
import jax.numpy as jnp

from autopdex import solution_structures



### Linear equations

def transport_equation(c):
  """
  Transport equation: du/dt + c * du/dx = 0

  Parameters:
    c (float): Transport coefficient.

  Returns:
    pde_fun (function): Function to compute the PDE residual.
  """
  def pde_fun(x, ansatz, settings, static_settings, int_point_number, set):
    u = ansatz
    u_derivative = jax.jacrev(u)(x)
    u_x = u_derivative[0]
    u_t = u_derivative[1]
    return u_t + c * u_x
  return pde_fun

def poisson(coefficient_fun = lambda x, settings: 1., source_fun=None):
  """
  Poisson equation in n dimensions: coefficient * Laplace(Theta) + source_term = 0

  Parameters:
    coefficient_fun (function): Function to compute the coefficient
    source_fun (function (dependend on the position jnp.ndarray x), optional): Function to compute the source term, defaults to None.

  Returns:
    pde_fun (function): Function to compute the PDE residual.
  """

  def pde_fun(x, ansatz, settings, static_settings, int_point_number, set):
    hessian = jax.jacfwd(jax.jacrev(ansatz))(x)

    # Coefficient
    coefficient = coefficient_fun(x, settings)

    if source_fun==None:
      pde = coefficient * jnp.trace(hessian)
    else:
      pde = coefficient * jnp.trace(hessian) + source_fun(x)

    return pde
  return pde_fun

def poisson_weak(coefficient_fun = lambda x, settings: 1., source_fun=None):
  """
  Poisson equation in n dimensions as weak form.

  Parameters:
    coefficient_fun (function): Function to compute the coefficient of the laplacian, defaults to 1.
    source_fun (function (dependend on the position jnp.ndarray x), optional): Function to compute the source term, defaults to None.

  Returns:
    pde_fun (function): Function to compute the weak form of the PDE.
  """

  def pde_fun(x, ansatz, test_ansatz, settings, static_settings, int_point_number, set):
    # Virtual temperature change and gradient
    var_theta = test_ansatz(x)
    var_theta_grad = jax.jacrev(test_ansatz)(x)

    # Temperature gradient
    theta_grad = jax.jacrev(ansatz)(x)

    # Coefficient
    coefficient = coefficient_fun(x, settings)

    if source_fun==None:
      pde = -coefficient * jnp.dot(theta_grad, var_theta_grad)
    else:
      pde = -coefficient * jnp.dot(theta_grad, var_theta_grad) + source_fun(x) * var_theta

    return pde
  return pde_fun

def poisson_fos(spacing, coefficient_fun = lambda x, settings: 1., source_fun=None):
  """
  Poisson equation as set of first order models augmented by curl(v)=0
  
  First field of solution space has to be scalar function; remaining fields form the gradient, i.e.
  theta = lambda x: ansatz(x)[0]
  grad_theta = lambda x: ansatz(x)[1:]
  n_dim = 2 or 3

  Parameters:
    spacing (float): Spacing for the weights.
    coefficient_fun (function): Function to compute the coefficient of the laplacian, defaults to 1.
    source_fun (function, optional): Function to compute the source term, defaults to None.

  Returns:
    pde_fun (function): Function to compute the first order PDE system residuals.
  """

  def pde_fun(x, ansatz, settings, static_settings, int_point_number, set):
    n_dim = x.shape[0]
    
    # Coefficient
    coefficient = coefficient_fun(x, settings)

    # Primary fields
    theta = lambda t: ansatz(t)[0]
    grad_theta = lambda t: ansatz(t)[1:] # v
    grad_v = jax.jacfwd(grad_theta)(x)

    # Poisson equation
    if source_fun==None:
      pde_1 = coefficient * jnp.asarray([jnp.trace(grad_v)])
    else:
      pde_1 = coefficient * jnp.asarray([jnp.trace(grad_v) + source_fun(x)])

    # Substituted derivatives
    pde_2 = grad_theta(x) - jax.jacrev(theta)(x)

    # Additional constraint in order to make functional fully H1 coercive
    if n_dim == 2:
      curl_v = jnp.asarray([grad_v[1,0] - grad_v[0,1]])
    elif n_dim == 3:
      curl_v = jnp.asarray([grad_v[2,1] - grad_v[1,2],
                            grad_v[0,2] - grad_v[2,0],
                            grad_v[1,0] - grad_v[0,1]])
    else:
      assert False, "This dimensionality is not supported by laplacian_fos"
    pde_3 = curl_v

    w1 = 1
    w2 = 1 / spacing
    w3 = 1 / spacing
    return jnp.concatenate([w1*pde_1, w2*pde_2, w3*pde_3]).flatten()
  return pde_fun

def heat_equation(diffusivity_fun):
  """
  Space-time heat equation with thermal diffusivity alpha.

  Last dimension is the time dimension, remaining dimensions are spatial dimensions

  Parameters:
    diffusivity_fun (function): Function to compute the thermal diffusivity.

  Returns:
    heat_fun (function): Function to compute the PDE residual.
  """

  def heat_fun(x, ansatz, settings, static_settings, int_point_number, set):
    n_dim = x.shape[0]
    alpha = diffusivity_fun(x)

    rate = jax.jacfwd(ansatz)(x)[-1] # du / dt
    spatial_derivatives = jax.jacfwd(jax.jacrev(ansatz))(x).diagonal()[:n_dim-1]
    laplacian = spatial_derivatives.sum() # Laplace(u)

    pde = rate - alpha * laplacian
    return pde

  return heat_fun

def heat_equation_fos(diffusivity_fun):
  """Space-time heat equation with thermal diffusivity alpha as a first order system with curl-augmentation for more than 2 spatial dimensions
  
  - supports n_dim = 2 to n_dim = 3, i.e. 1 to 3 spatial dimensions plus one time dimension
  - number of fiels: n_dim (1 temperature field, n_dim-1 spatial gradients)
  - last dimension is the time dimension, the remaining ones are spacial dimensions

  e.g. n_dim = 2:
  - first dimension is x-dimension, second dimension is time dimension
  - first field is temperature, second field is derivative of temperature with respect to x

  Parameters:
    diffusivity_fun (function): Function to compute the thermal diffusivity.

  Returns:
    heat_fun (function): Function to compute the PDE residual.
  """
  
  def heat_fun(x, ansatz, settings, static_settings, int_point_number, set):
    n_dim = x.shape[0]
    alpha = diffusivity_fun(x)

    temperature = lambda t: ansatz(t)[0] # u
    derivative_field = lambda t: ansatz(t)[1:] # v
    derivatives = jax.jacrev(temperature)(x) # [du/dx, du/dt] (flat)

    # Heat equation with substituted spatial derivatives v
    rate = derivatives[-1] # du / dt
    grad_v = jax.jacfwd(derivative_field)(x)[:,:n_dim-1]
    laplacian = grad_v.diagonal().sum() # div(v)
    pde_1 = jnp.asarray([rate - alpha * laplacian])

    # Substituted spatial derivatives: v - du / dx
    pde_2 = derivative_field(x) - derivatives[:n_dim-1]

    # Augmentation for stabilization
    if n_dim == 2:
      curl_v = jnp.asarray([0.])
    elif n_dim == 3:
      curl_v = jnp.asarray([grad_v[1,0] - grad_v[0,1]])
    elif n_dim == 4:
      curl_v = jnp.asarray([grad_v[2,1] - grad_v[1,2],
                            grad_v[0,2] - grad_v[2,0],
                            grad_v[1,0] - grad_v[0,1]])
    else:
      assert False, "This dimensionality is not supported by heat_equation_fos"
    pde_3 = curl_v

    return jnp.concatenate([pde_1, pde_2, pde_3]).flatten()
  return heat_fun

def d_alembert(wave_number_fun):
  """
  Space-time d'Alembert operator 1 temporal [-1] and (n-1) spatial [:n-1] dimension with wave number c.

  Parameters:
    wave_number_fun (function): Function to compute the wave number.

  Returns:
    d_alembert_fun (function): Function to compute the PDE residual.
  """

  def d_alembert_fun(x, ansatz, settings, static_settings, int_point_number, set):
    c = wave_number_fun(x)

    hessian = jax.jacfwd(jax.jacrev(ansatz))(x)
    hessian_diagonal = jnp.diagonal()
    coefficients = - jnp.ones_like(hessian_diagonal)
    coefficients = coefficients.at[-1].set(1 / c**2)

    pde = jnp.dot(hessian_diagonal, coefficients)
    return pde
  
  return d_alembert_fun

def d_alembert_fos(wave_number_fun, spacing):
  """
  Constructs a first-order system for the space-time d'Alembert operator with wave speed c.

  This function returns a PDE function representing the d'Alembert operator for one temporal dimension 
  and (n-1) spatial dimensions, with augmented conditions to ensure a fully H1 coercive functional.

  Parameters:
    wave_number_fun (function): A function that computes the wave number (inverse of wave speed) given spatial coordinates.
    spacing (float): The spatial spacing parameter used to scale the augmented conditions.

  Returns:
    function: A function that evaluates the first-order system of PDEs for the d'Alembert operator, which includes:
      - The primary d'Alembert operator.
      - Substituted derivative fields.
      - Additional constraints for stabilization (curl conditions).
  
  Notes:
  - The last dimension of the input coordinates x is considered the time dimension; the remaining dimensions are spatial dimensions.
  """

  def d_alembert_fun(x, ansatz, settings, static_settings, int_point_number, set):
    # Initialization
    n_dim = x.shape[0]
    c = wave_number_fun(x)
    field = lambda t: ansatz(t)[0]
    derivative_field = lambda t: ansatz(t)[1:]
    
    # First pde: D' Alembert operator
    hessian = jax.jacfwd(derivative_field)(x)
    hessian_diagonal = hessian.diagonal()
    coefficients = - jnp.ones_like(hessian_diagonal)
    coefficients = coefficients.at[0].set(1 / c) # The other c is already in the other pde
    pde_1 = jnp.asarray([jnp.dot(hessian_diagonal, coefficients)])

    # Second pde: Gradient field
    scaled_grad = jax.jacrev(field)(x)
    scaled_grad = scaled_grad.at[-1].multiply(1 / c)
    pde_2 = derivative_field(x) - scaled_grad

    # Additional constraint in order to make functional fully H1 coercive
    # grad_v = derivative_field(x)
    grad_v = jax.jacfwd(derivative_field)(x)
    if n_dim == 2:
      curl_v = jnp.asarray([0.])
    elif n_dim == 3:
      curl_v = jnp.asarray([grad_v[1,0] - grad_v[0,1]])
    elif n_dim == 4:
      curl_v = jnp.asarray([grad_v[2,1] - grad_v[1,2],
                            grad_v[0,2] - grad_v[2,0],
                            grad_v[1,0] - grad_v[0,1]])
    else:
      assert False, "This dimensionality is not supported by laplacian_fos"
    pde_3 = curl_v

    w1 = 1
    w2 = 1 / spacing
    w3 = 1 / spacing
    return jnp.concatenate([w1*pde_1, w2*pde_2, w3*pde_3]).flatten()

  return d_alembert_fun

def linear_elasticity(youngs_mod_fun, poisson_ratio_fun, mode, volume_load_fun=None):
  """
  Constructs the strong form of linear elasticity in Voigt notation, displacement based.

  This function supports 2D plain strain, 2D plain stress, and 3D elasticity problems.

  Parameters:
  youngs_mod_fun (function): Function to compute Young's modulus given spatial coordinates (and optional settings).
  poisson_ratio_fun (function): Function to compute Poisson's ratio given spatial coordinates (and optional settings).
  mode (str): Specifies the mode of the elasticity problem. It can be 'plain strain', 'plain stress', or '3d'.
  volume_load_fun (function, optional): Function to compute volume load given spatial coordinates. Default is None.

  Returns:
  function: A function that evaluates the strong form PDE for linear elasticity.
  """
  if mode == 'plain strain' or mode == 'plain stress':
    n_dim = 2
  elif mode == '3d':
    n_dim = 3
  else:
    assert False, '\'mode\' for linear elasticity not properly set.'

  def lin_el_fun(x, ansatz, settings, static_settings, int_point_number, set):
    def get_stress(xf):
      # Kinematics
      displacement = ansatz
      displ_grad = jax.jacfwd(ansatz)(xf)
      strain = (1/2) * (displ_grad + displ_grad.transpose())
      if n_dim == 2:
        strain_voigt = jnp.asarray([strain[0,0], strain[1,1], 2*strain[0,1]])
      else:
        strain_voigt = jnp.asarray([strain[0,0], strain[1,1], strain[2,2], 2*strain[0,1], 2*strain[0,2], 2*strain[1,2]])

      # Constiutive model
      if len(signature(poisson_ratio_fun).parameters) == 1:
        nu = poisson_ratio_fun(x)
      else:
        nu = poisson_ratio_fun(x, settings)
      if len(signature(youngs_mod_fun).parameters) == 1:
        Em = youngs_mod_fun(x)
      else:
        Em = youngs_mod_fun(x, settings)

      match mode:
        case 'plain strain':
          mu = Em / (2 * (1 + nu))
          c1 = 1 - 2 * nu
          c2 = 1 - nu
          coeff = 2 * mu / c1
          material_tangent = coeff * jnp.asarray([[c2, nu, 0],
                                                  [nu, c2, 0],
                                                  [0, 0, c1]])

        case 'plain stress':
          coeff = Em / (1 - nu**2)
          material_tangent = coeff * jnp.asarray([[1, nu, 0],
                                                  [nu, 1, 0],
                                                  [0, 0, (1 - nu) / 2]])

        case '3d':
          coeff = Em / (1 + nu)
          c1 = 1 - 2 * nu
          c2 = (1 - nu) / c1
          c3 = nu / c1
          c4 = 1 / 2
          material_tangent = coeff * jnp.asarray([[c2, c3, c3, 0, 0, 0],
                                                  [c3, c2, c3, 0, 0, 0],
                                                  [c3, c3, c2, 0, 0, 0],
                                                  [0, 0, 0, c4, 0, 0],
                                                  [0, 0, 0, 0, c4, 0],
                                                  [0, 0, 0, 0, 0, c4]])

      return material_tangent @ strain_voigt
    # stress_voigt = get_stress(x)

    # Internal forces
    grad_stress = jax.jacfwd(get_stress)(x)
    if n_dim == 2:
      div_stress = jnp.asarray([grad_stress[0,0] + grad_stress[2,1],
                                grad_stress[2,0] + grad_stress[1,1]])
    else:
      div_stress = jnp.asarray([grad_stress[0,0] + grad_stress[3,1] + grad_stress[4,2],
                                grad_stress[3,0] + grad_stress[1,1] + grad_stress[5,2],
                                grad_stress[4,0] + grad_stress[5,1] + grad_stress[2,2]])

    # Balance equation
    if volume_load_fun == None:
      return div_stress
    else:
      return div_stress + volume_load_fun(x)
  return lin_el_fun

def linear_elasticity_weak(youngs_mod_fun, poisson_ratio_fun, mode, volume_load_fun=None):
  """ 
  Constructs the weak form of linear elasticity in Voigt notation, displacement based.

  This function supports 2D plain strain, 2D plain stress, and 3D elasticity problems.

  Parameters:
  youngs_mod_fun (function): Function to compute Young's modulus given spatial coordinates (and optional settings).
  poisson_ratio_fun (function): Function to compute Poisson's ratio given spatial coordinates (and optional settings).
  mode (str): Specifies the mode of the elasticity problem. It can be 'plain strain', 'plain stress', or '3d'.
  volume_load_fun (function, optional): Function to compute volume load given spatial coordinates. Default is None.

  Returns:
  function: A function that evaluates the weak form PDE for linear elasticity.

  Todo: 
  - anisotropy modes
  - support for dynamic problems
  """
  if mode == 'plain strain' or mode == 'plain stress':
    n_dim = 2
  elif mode == '3d':
    n_dim = 3
  else:
    assert False, '\'mode\' for linear elasticity not properly set.'

  def lin_el_fun(x, ansatz, test_ansatz, settings, static_settings, int_point_number, set):
    def get_stress(xf):
      # Kinematics
      displacement = ansatz
      displ_grad = jax.jacfwd(ansatz)(xf)
      strain = (1/2) * (displ_grad + displ_grad.transpose())
      if n_dim == 2:
        strain_voigt = jnp.asarray([strain[0,0], strain[1,1], 2*strain[0,1]])
      else:
        strain_voigt = jnp.asarray([strain[0,0], strain[1,1], strain[2,2], 2*strain[0,1], 2*strain[0,2], 2*strain[1,2]])

      # Constiutive model
      if len(signature(poisson_ratio_fun).parameters) == 1:
        nu = poisson_ratio_fun(x)
      else:
        nu = poisson_ratio_fun(x, settings)
      if len(signature(youngs_mod_fun).parameters) == 1:
        Em = youngs_mod_fun(x)
      else:
        Em = youngs_mod_fun(x, settings)
      match mode:
        case 'plain strain':
          mu = Em / (2 * (1 + nu))
          c1 = 1 - 2 * nu
          c2 = 1 - nu
          coeff = 2 * mu / c1
          material_tangent = coeff * jnp.asarray([[c2, nu, 0],
                                                  [nu, c2, 0],
                                                  [0, 0, c1]])

        case 'plain stress':
          coeff = Em / (1 - nu**2)
          material_tangent = coeff * jnp.asarray([[1, nu, 0],
                                                  [nu, 1, 0],
                                                  [0, 0, (1 - nu) / 2]])

        case '3d':
          coeff = Em / (1 + nu)
          c1 = 1 - 2 * nu
          c2 = (1 - nu) / c1
          c3 = nu / c1
          c4 = 1 / 2
          material_tangent = coeff * jnp.asarray([[c2, c3, c3, 0, 0, 0],
                                                  [c3, c2, c3, 0, 0, 0],
                                                  [c3, c3, c2, 0, 0, 0],
                                                  [0, 0, 0, c4, 0, 0],
                                                  [0, 0, 0, 0, c4, 0],
                                                  [0, 0, 0, 0, 0, c4]])

      return material_tangent @ strain_voigt
    stress_voigt = get_stress(x)

    # Test function
    virtual_disp = test_ansatz(x)
    virtual_disp_grad = jax.jacfwd(test_ansatz)(x)
    virtual_strain = (1/2) * (virtual_disp_grad + virtual_disp_grad.transpose())
    if n_dim == 2:
      virtual_strain_voigt = jnp.asarray([virtual_strain[0,0], virtual_strain[1,1], 2*virtual_strain[0,1]])
    else:
      virtual_strain_voigt = jnp.asarray([virtual_strain[0,0], virtual_strain[1,1], virtual_strain[2,2], 2*virtual_strain[0,1], 2*virtual_strain[0,2], 2*virtual_strain[1,2]])

    # Balance equation
    if volume_load_fun == None:
      return jnp.dot(stress_voigt, virtual_strain_voigt)
    else:
      return jnp.dot(stress_voigt, virtual_strain_voigt) - jnp.dot(volume_load_fun(x), virtual_disp)
  return lin_el_fun

def linear_elasticity_fos(youngs_mod_fun, poisson_ratio_fun, mode, spacing, volume_load_fun=None):
  """
  Constructs the first order system (FOS) of linear elasticity in Voigt notation, displacement based.

  This function supports 2D plain strain and 3D elasticity problems. The formulation is based on the first order 
  system described in https://doi.org/10.1137/S0036142902418357 with additional asymmetry penalization.

  Parameters:
    youngs_mod_fun (function): Function to compute Young's modulus given spatial coordinates (and optional settings).
    poisson_ratio_fun (function): Function to compute Poisson's ratio given spatial coordinates (and optional settings).
    mode (str): Specifies the mode of the elasticity problem. It can be 'plain strain' or '3d'.
    spacing (float): The spatial spacing parameter used to scale the augmented conditions.
    volume_load_fun (function, optional): Function to compute volume load given spatial coordinates. Default is None.

  Returns:
    function: A function that evaluates the first order system of PDEs for linear elasticity.

  The returned function computes the following:
    - The kinematic strain based on the displacement gradient.
    - The constitutive relation using the specified Young's modulus and Poisson's ratio.
    - The internal force balance equation, optionally including volume loads.
    - Weak enforcement of symmetry of the Cauchy stress tensor.

  Notes:
    - In 2D, the number of fields is 2 (displacement) + 4 (stress) = 6.
    - In 3D, the number of fields is 3 (displacement) + 9 (stress) = 12.
    - This function only supports 'plain strain' and '3d' modes.
  """
  if mode == 'plain strain':
    n_dim = 2
  elif mode == '3d':
    n_dim = 3
  else:
    assert False, '\'mode\' for linear elasticity not properly set.'

  def lin_el_fun(x, ansatz, settings, static_settings, int_point_number, set):
    # Primary fields
    displacement = lambda t: ansatz(t)[:n_dim]
    stress = lambda t: ansatz(t)[n_dim:].reshape((n_dim, n_dim))

    # Kinematics
    displ_grad = jax.jacfwd(displacement)(x)
    strain = (1/2) * (displ_grad + displ_grad.transpose())

    # Constiutive model
    if len(signature(poisson_ratio_fun).parameters) == 1:
      nu = poisson_ratio_fun(x)
    else:
      nu = poisson_ratio_fun(x, settings)
    if len(signature(youngs_mod_fun).parameters) == 1:
      Em = youngs_mod_fun(x)
    else:
      Em = youngs_mod_fun(x, settings)
    mu = Em / (2 * (1 + nu)) # Shear modulus
    lam = Em * nu / ((1 + nu) * (1 - 2 * nu))# First Lame constant

    # Complience tensor times stress
    sigma = stress(x)
    A_sigma = (1 / (2 * mu)) * (sigma - (lam / (n_dim * lam + 2 * mu)) * jnp.trace(sigma) * jnp.identity(n_dim))

    # Second pde: constitutive model
    pde_2 = (A_sigma - strain).flatten()

    # Internal forces
    grad_stress = jax.jacfwd(stress)(x)
    if n_dim == 2:
      div_stress = jnp.asarray([grad_stress[0,0,0] + grad_stress[0,1,1],
                                grad_stress[1,0,0] + grad_stress[1,1,1]])
    else:
      div_stress = jnp.asarray([grad_stress[0,0,0] + grad_stress[0,1,1] + grad_stress[0,2,2],
                                grad_stress[1,0,0] + grad_stress[1,1,1] + grad_stress[1,2,2],
                                grad_stress[2,0,0] + grad_stress[2,1,1] + grad_stress[2,2,2]])

    # First pde: Balance equation
    if volume_load_fun == None:
      pde_1 = div_stress
    else:
      pde_1 = div_stress + volume_load_fun(x)
    
    # Weak enforcement of symmetry of Cauchy stress
    pde_3 = ((1/2) * (stress(x) - stress(x).transpose())).flatten()

    # Return first order system
    w_1 = 1.0
    w_2 = 1.0
    w_3 = 10.0 / mu
    return jnp.concatenate([w_1*pde_1, w_2*pde_2, w_3*pde_3]).flatten()
  
  # Return function for evaluating first order system with arguments x and ansatz
  return lin_el_fun

def neumann_weak(neumann_fun):
  """
  Constructs the weak form of Neumann boundary conditions for the virtual work of surface tractions or heat inflow.

  This function can be used to impose Neumann boundary conditions in the weak form of problems such as linear 
  momentum balance or heat conduction.

  Parameters:
    neumann_fun (callable): A function that computes the Neumann boundary condition (e.g., surface tractions or heat inflow)
        given spatial coordinates. This function can optionally take the settings as an additional argument.

  Returns:
    function
        A function that evaluates the weak form of the Neumann boundary condition.

  The returned function computes the virtual work of the Neumann boundary condition, which is the dot product of 
  the virtual displacement (or temperature) and the Neumann boundary condition (e.g., surface tractions or heat inflow).
  """

  def fun(x, ansatz, test_ansatz, settings, static_settings, int_point_number, set):
    # Test function
    virtual_disp = test_ansatz(x)

    # Tractions
    if len(signature(neumann_fun).parameters) == 1:
      tractions = neumann_fun(x)
    else:
      tractions = neumann_fun(x, settings)

    # Virtual work
    return - jnp.dot(virtual_disp, tractions)
  return fun


### Nonlinear equations

def burgers_equation_inviscid():
  """
  Constructs the inviscid Burgers' equation: du/dt + u * du/dx = 0.

  Parameters:
    None

  Returns:
    function: A function that evaluates the inviscid Burgers' equation.

  Notes:
    - first dimension: x
    - second dimension: t
  """
  def pde_fun(x, ansatz, settings, static_settings, int_point_number, set):
    u = ansatz(x)
    u_x = jax.jacrev(ansatz)(x)[0]
    u_t = jax.jacrev(ansatz)(x)[1]
    return u_t + u * u_x
  return pde_fun

def hyperelastic_steady_state_fos(strain_energy_fun, youngs_mod_fun, poisson_ratio_fun, mode, spacing, volume_load_fun=None):
  """
  Constructs a first-order PDE system for given strain energy function

  This function is based on equation (28) in https://doi.org/10.1002/nme.5951 and returns a function 
  representing the first-order PDE system for steady-state hyperelasticity.

  Parameters:
    strain_energy_fun (function): Function that returns the strain energy given the deformation gradient and Lame parameters.
    youngs_mod_fun (function): Function to compute Young's modulus given spatial coordinates.
    poisson_ratio_fun (function): Function to compute Poisson's ratio given spatial coordinates.
    mode (str): Specifies the mode of the elasticity problem. It can be 'plain strain' or '3d'.
    spacing (float): The spatial spacing parameter used to scale the augmented conditions.
    volume_load_fun (function, optional): Function to compute volume load given spatial coordinates. Default is None.

  Returns:
    function: A function that evaluates the first-order PDE system for hyperelasticity.

  Notes:
    - The volume load function, if provided, must be volume-specific.
  """

  def pde_fun(x, ansatz, settings, static_settings, int_point_number, set):
    # Initialization
    n_dim = x.shape[0]
    if len(signature(poisson_ratio_fun).parameters) == 1:
      nu = poisson_ratio_fun(x)
    else:
      nu = poisson_ratio_fun(x, settings)
    if len(signature(youngs_mod_fun).parameters) == 1:
      Em = youngs_mod_fun(x)
    else:
      Em = youngs_mod_fun(x, settings)

    # Lame parameters
    lam = Em * nu / ((1 + nu) * (1 - 2 * nu))
    mu = Em / (2 * (1 + nu))

    # Ansatz
    u = lambda t: ansatz(t)[:n_dim] # Displacement
    P = lambda t: ansatz(t)[n_dim:].reshape((n_dim, n_dim)) # 1. Piola-Kirchoff stress

    # Linear momentum balance
    Grad_P = jax.jacfwd(P)(x)
    Div_P = jnp.trace(Grad_P, axis1=1, axis2=2)

    # First pde: Balance equation
    if volume_load_fun == None:
      pde_1 = Div_P
    else:
      b = volume_load_fun(x)
      pde_1 = Div_P + b

    if mode == 'plain strain':
      # Deformation gradient
      H = jax.jacfwd(u)(x)
      F = jnp.asarray([jnp.asarray([H[0,0], H[0,1], 0.]),
                       jnp.asarray([H[1,0], H[1,1], 0.]),
                       jnp.asarray([0., 0., 0.])]) + jnp.identity(3)
      
      # First Piola-Kirchhoff stress
      dpsi_dF = jax.jacrev(neo_hooke)(F, (lam, mu))
      dpsi_dF_FT = (dpsi_dF @ jnp.transpose(F))[:n_dim,:n_dim]
      PFT = P(x)[:n_dim,:n_dim] @ jnp.transpose(F[:n_dim,:n_dim])

    elif mode == '3d':
      # Deformation gradient
      F = jax.jacfwd(u)(x) + jnp.identity(3)


      # First Piola-Kirchhoff stress
      dpsi_dF = jax.jacrev(strain_energy_fun)(F, (lam, mu))
      dpsi_dF_FT = dpsi_dF @ jnp.transpose(F)
      PFT = P(x) @ F.transpose()

    else:
      assert False, 'Hyperelastic model supports only \'plain strain\' and \'3d\' modes.'

    # Constitutive law
    pde_2 = ((1/2) * (PFT + jnp.transpose(PFT)) - dpsi_dF_FT).flatten()

    # Enforce symmetry of P \cdot F^T
    PFT_asym = (1/2) * (PFT - PFT.transpose())
    pde_3 = PFT_asym.flatten()

    # Weightings
    w_1 = 1.0
    w_2 = 1.0 / mu
    w_3 = 10.0 / mu
    return jnp.concatenate([w_1 * pde_1, w_2 * pde_2, w_3 * pde_3])
  return pde_fun

def hyperelastic_steady_state_weak(strain_energy_fun, youngs_mod_fun, poisson_ratio_fun, mode, volume_load_fun=None):
  """
  Constructs the weak form of hyperelasticity for given strain energy function

  This function returns a function representing the weak form PDE system for steady-state hyperelasticity.

  Parameters:
    strain_energy_fun (function): Function that returns the strain energy given the deformation gradient and Lame parameters.
    youngs_mod_fun (function): Function to compute Young's modulus given spatial coordinates.
    poisson_ratio_fun (function): Function to compute Poisson's ratio given spatial coordinates.
    mode (str): Specifies the mode of the elasticity problem. It can be 'plain strain' or '3d'.
    volume_load_fun (function, optional): Function to compute volume load given spatial coordinates. Default is None.

  Returns:
    function: A function that evaluates the weak form PDE system for hyperelasticity.

  Notes:
    - The volume load function, if provided, must be volume-specific.
  """

  def pde_fun(x, ansatz, test_ansatz, settings, static_settings, int_point_number, set):
    # Initialization
    n_dim = x.shape[0]
    if len(signature(poisson_ratio_fun).parameters) == 1:
      nu = poisson_ratio_fun(x)
    else:
      nu = poisson_ratio_fun(x, settings)
    if len(signature(youngs_mod_fun).parameters) == 1:
      Em = youngs_mod_fun(x)
    else:
      Em = youngs_mod_fun(x, settings)

    # Lame parameters
    lam = Em * nu / ((1 + nu) * (1 - 2 * nu))
    mu = Em / (2 * (1 + nu))

    # (Virtual) displacement
    u = ansatz
    virt_u = test_ansatz

    # Deformation gradient
    if mode == 'plain strain':
      H = jax.jacfwd(u)(x)
      F = jnp.asarray([[H[0,0], H[0,1], 0.],
                       [H[1,0], H[1,1], 0.],
                       [0.,     0.,     0.]]) + jnp.identity(3)
      
      virt_H = jax.jacfwd(virt_u)(x)
      virt_F = jnp.asarray([[virt_H[0,0], virt_H[0,1],  0.],
                            [virt_H[1,0], virt_H[1,1],  0.],
                            [0.,          0.,           0.]])
    elif mode == '3d':
      F = jax.jacfwd(u)(x) + jnp.identity(3)
      virt_F = jax.jacfwd(virt_u)(x)
    else:
      assert False, 'Hyperelastic model supports only \'plain strain\' and \'3d\' modes.'

    # First Piola-Kirchhoff stress
    P = jax.jacrev(strain_energy_fun)(F, (lam, mu))

    # Virtual work of internal forces
    PI_int = jnp.einsum('ij, ij -> ', P, virt_F)
    # PI_int = jnp.trace(P @ virt_F.transpose())

    if volume_load_fun == None:
      return PI_int
    else:
      # Add virtual work of volume forces
      b = volume_load_fun(x)
      PI_ext = jnp.dot(b, virt_u)
      return PI_int - PI_ext
  return pde_fun

def navier_stokes_incompressible_steady(dynamic_viscosity, density=1.0, incompressible_weighting=1., volume_load_fun=None):
  """
  Constructs the steady-state incompressible Navier-Stokes equations.

  Parameters:
    dynamic_viscosity (float): Dynamic viscosity of the fluid.
    density (float, optional): Density of the fluid. Default is 1.0.
    incompressible_weighting (float, optional): Weighting factor for the incompressibility constraint. Default is 1.0.
    volume_load_fun (function, optional): Function to compute volume load given spatial coordinates. Default is None.

  Returns:
    function: A function that evaluates the steady-state incompressible Navier-Stokes equations.

  Notes:
    - The first equation is the constraint of a divergence-free flow field.
    - The remaining equations are the linear momentum balance.
    - The first field in the ansatz function is the pressure field, and the remaining fields are the velocity components.
    - The density and viscosity are assumed to be constant.
    - The volume load function, if provided, must be volume-specific.
  """

  def pde_fun(x, ansatz, settings, static_settings, int_point_number, set):
    pressure = lambda t: ansatz(t)[0] # p
    velocity = lambda t: ansatz(t)[1:] # v

    grad_p = jax.jacfwd(pressure)(x)
    grad_v = jax.jacfwd(velocity)(x)
    laplace_v = jnp.trace(jax.jacfwd(jax.jacfwd(velocity))(x), axis1=1, axis2=2)

    # Incompressible flow: div(v) = 0
    pde_1 = jnp.asarray([jnp.trace(grad_v)])

    # Linear momentum balance
    if volume_load_fun==None:
      pde_2 = grad_v @ velocity(x) + (1/density) * grad_p - (dynamic_viscosity / density) * laplace_v * jnp.ones_like(velocity(x))
    else:
      pde_2 = grad_v @ velocity(x) + (1/density) * grad_p - (dynamic_viscosity / density) * laplace_v * jnp.ones_like(velocity(x)) - volume_load_fun(x)
    
    return jnp.concatenate([incompressible_weighting * pde_1, pde_2]).flatten()
  return pde_fun

def navier_stokes_incompressible(dynamic_viscosity, density=1.0, incompressible_weighting=1., volume_load_fun=None):
  """
  Constructs the transient incompressible Navier-Stokes equations.

  Parameters:
    dynamic_viscosity (float): Dynamic viscosity of the fluid.
    density (float, optional): Density of the fluid. Default is 1.0.
    incompressible_weighting (float, optional): Weighting factor for the incompressibility constraint. Default is 1.0.
    volume_load_fun (function, optional): Function to compute volume load given spatial coordinates. Default is None.

  Returns:
    function: A function that evaluates the transient incompressible Navier-Stokes equations.

  Notes:
    - The first equation is the constraint of a divergence-free flow field.
    - The remaining equations are the linear momentum balance.
    - The first field in the ansatz function is the pressure field, and the remaining fields are the velocity components.
    - The density and viscosity are assumed to be constant.
    - The volume load function, if provided, must be volume-specific.
  """

  def pde_fun(x, ansatz, settings, static_settings, int_point_number, set):
    # Number of spatial dimensions
    n_sdim = x.shape[0]-1

    pressure = lambda t: ansatz(t)[0] # p
    velocity = lambda t: ansatz(t)[1:] # v

    grad_p = jax.jacfwd(pressure)(x)[:n_sdim]
    grad_v = jax.jacfwd(velocity)(x)[:,:n_sdim]
    dvdt = jax.jacfwd(velocity)(x)[:,n_sdim]
    laplace_v = jnp.trace(jax.jacfwd(jax.jacfwd(velocity))(x)[:,:,:n_sdim], axis1=1, axis2=2)

    # Incompressible flow: div(v) = 0
    pde_1 = jnp.asarray([jnp.trace(grad_v)])

    # Linear momentum balance
    if volume_load_fun==None:
      pde_2 = dvdt + grad_v @ velocity(x) + (1/density) * grad_p - (dynamic_viscosity / density) * laplace_v * jnp.ones_like(velocity(x))
    else:
      pde_2 = dvdt + grad_v @ velocity(x) + (1/density) * grad_p - (dynamic_viscosity / density) * laplace_v * jnp.ones_like(velocity(x)) - volume_load_fun(x)
    
    return jnp.concatenate([incompressible_weighting * pde_1, pde_2]).flatten()
  return pde_fun


### Strain energy functions

def neo_hooke(F, param):
  """
  Computes the strain energy for a neo-Hookean material model of Ciarlet type given the deformation gradient.

  Parameters:
    F (jnp.ndarray): Deformation gradient tensor.
    param (tuple): Lame parameters (lambda, mu).

  Returns:
    float: Strain energy value for the given deformation gradient.

  Notes:
    - The strain energy function is based on the Ciarlet type neo-Hookean model.
    - This model is suitable for large deformation analysis in hyperelastic materials.
  """
  (lam, mu) = param

  # Right Cauchy-Green tensor
  tr_C = jnp.einsum('ij, ij', F, F)
  J = jnp.linalg.det(F)
  ln_J = jnp.log(J)

  # Neo-Hookean strain energy function of Ciarlet type
  psi = (mu/2) * (tr_C - 3 - 2*ln_J) + (lam/4) * (J**2 - 1 - 2*ln_J)
  return psi

def linear_elastic_strain_energy(F, param):
  """
  Computes the strain energy for a linear elastic material for small deformations.

  Parameters:
    F (jnp.ndarray): Deformation gradient tensor.
    param (tuple): Lame parameters (lambda, mu).

  Returns:
    float: Strain energy value for the given deformation gradient.
  """
  (lam, mu) = param

  # Right Cauchy-Green tensor
  disp_grad = F - jnp.identity(F.shape[0])
  strain = (disp_grad + disp_grad.T) / 2

  return (lam / 2) * (jnp.trace(strain))**2 + mu * jnp.trace(strain @ strain.T)


### User potentials/elements
def isoparametric_domain_integrate_potential(integrand_fun, ansatz_fun, ref_int_coor, ref_int_weights, initial_config=True):
  """
  Constructs a local integrand fun for integration of functions in the reference configuration of isoparametric elements.

  Parameters:
    integrand_fun (callable): Function that evaluates the integrand given the integration point, trial ansatz, settings, static settings, and element number.
    ansatz_fun (callable): Function that constructs the ansatz (trial and test) functions.
    ref_int_coor (jnp.ndarray): Reference integration coordinates for the isoparametric element.
    ref_int_weights (jnp.ndarray): Reference integration weights for the isoparametric element.
    initial_config (bool, optional): If True, use the initial configuration for isoparametric mapping. Default is True.

  Returns:
    function
        A function that evaluates the integrand for the isoparametric element.

  The returned function has the following parameters:
    - fI (jnp.ndarray): Element nodal values.
    - xI (jnp.ndarray): Element nodal coordinates.
    - elem_number (int): Element number.
    - settings (dict): Settings for the computation.
    - static_settings (dict): Static settings for the computation.
  """
  def user_element(fI, xI, elem_number, settings, static_settings, set):
    n_dofs = fI.flatten().shape[0]
    n_dim = xI.shape[-1]

    # One Gauß point contribution
    def functional(x_int, w_int):

      def integrated_weak_form(trial_fI):
        # Define trial and test function ansatz
        trial_ansatz = lambda x: ansatz_fun(x, xI, trial_fI, settings, True, n_dim)

        # Evaluate functional
        discrete_functional_fun = integrand_fun(x_int, trial_ansatz, settings, static_settings, elem_number, set)

        # Use ansatz for isoparametric mapping and compute determinant of Jacobian
        if initial_config:
          mapping = lambda x: ansatz_fun(x, xI, xI, settings, False, n_dim)
        else:
          # Take first n_dim fields as displacements
          uI = fI[:, :xI.shape[-1]]
          mapping = lambda x: ansatz_fun(x, xI + uI, xI + uI, settings, overwrite_diff=False)

        jacobian = jax.jacfwd(mapping)(x_int)
        det_jacobian = jnp.linalg.det(jacobian) 

        return w_int * det_jacobian * discrete_functional_fun

      # Residual
      return integrated_weak_form(fI)

    # Sum over Gauß points
    functional_contrib = jax.vmap(functional, (0, 0), 0)(ref_int_coor, ref_int_weights)
    return functional_contrib.sum(axis=0)
  return user_element

def isoparametric_domain_element_galerkin(weak_form_fun, ansatz_fun, ref_int_coor, ref_int_weights, initial_config=True):
  """
  Constructs an isoparametric domain element for Galerkin methods using a given weak form function.

  This function returns a user element function that evaluates the residual and tangent matrix for an isoparametric 
  element based on the provided weak form and ansatz functions.
  The function can e.g. be used in combination with the weak_form_fun hyperelastic_steady_state_weak.

  Parameters:
    weak_form_fun (callable): Function that evaluates the weak form given the integration point, trial ansatz, 
        test ansatz, settings, static settings, and element number.
    ansatz_fun (callable): Function that constructs the ansatz (trial and test) functions.
    ref_int_coor (jnp.ndarray): Reference integration coordinates for the isoparametric element.
    ref_int_weights (jnp.ndarray): Reference integration weights for the isoparametric element.
    initial_config (bool, optional): If True, use the initial configuration for isoparametric mapping. Default is True.

  Returns:
    function
        A function that evaluates the residual and tangent matrix for the isoparametric element.

  Notes:
    - The returned function can be used to evaluate the residual and tangent matrix for the element.
    - The mode parameter in the returned function can be 'residual' or 'tangent'.

  The returned function has the following parameters:
    - fI (jnp.ndarray): Element nodal values.
    - xI (jnp.ndarray): Element nodal coordinates.
    - elem_number (int): Element number.
    - settings (dict): Settings for the computation.
    - static_settings (dict): Static settings for the computation.
    - mode (str): Specifies whether to compute the 'residual' or 'tangent'.
  """
  def user_element(fI, xI, elem_number, settings, static_settings, mode, set):
    n_dofs = fI.flatten().shape[0]
    n_dim = xI.shape[-1]

    # One Gauß point contribution
    def residual_and_tangent(x_int, w_int):

      def integrated_weak_form(test_fI, trial_fI):
        # Define trial and test function ansatz
        trial_ansatz = lambda x: ansatz_fun(x, xI, trial_fI, settings, True, n_dim)
        test_ansatz = lambda x: ansatz_fun(x, xI, test_fI, settings, True, n_dim)

        # Evaluate weak form
        discrete_weak_form = weak_form_fun(x_int, trial_ansatz, test_ansatz, settings, static_settings, elem_number, set)

        # Use ansatz for isoparametric mapping and compute determinant of Jacobian
        if initial_config:
          mapping = lambda x: ansatz_fun(x, xI, xI, settings, False, n_dim)
        else:
          # Take first n_dim fields as displacements
          uI = fI[:, :xI.shape[-1]]
          mapping = lambda x: ansatz_fun(x, xI + uI, xI + uI, settings, overwrite_diff=False)

        jacobian = jax.jacfwd(mapping)(x_int)
        det_jacobian = jnp.linalg.det(jacobian) 

        return w_int * det_jacobian * discrete_weak_form

      # Residual
      residual_fun = jax.jacrev(integrated_weak_form, argnums=0)
      residual = residual_fun(fI, fI).flatten()

      # Tangent
      tangent_fun = jax.jacfwd(residual_fun, argnums=1)
      tangent = tangent_fun(fI, fI).reshape(n_dofs,n_dofs)

      return (residual, tangent)

    # Sum over Gauß points
    (residual_contrib, tangent_contrib) = jax.vmap(residual_and_tangent, (0, 0), (0, 0))(ref_int_coor, ref_int_weights)
    residual = residual_contrib.sum(axis=0)
    tangent = tangent_contrib.sum(axis=0)

    if mode == 'residual':
      return residual
    elif mode =='tangent':
      return tangent
    else:
      assert False, 'User elem mode has to be \'residual\' or \'tangent\'.'

  return user_element

def isoparametric_surface_element_galerkin(weak_form_fun, ansatz_fun, ref_int_coor, ref_int_weights, tangent_contributions, initial_config=True):
  """
  Constructs an isoparametric surface element for Galerkin methods using a given weak form function.

  This function returns a user element function that evaluates the residual and tangent matrix for an isoparametric 
  surface element based on the provided weak form and ansatz functions. The function can, for example, be used in 
  combination with the weak_form_fun neumann_weak.

  Parameters:
    weak_form_fun (callable): Function that evaluates the weak form given the integration point, trial ansatz, 
        test ansatz, settings, static settings, and element number.
    ansatz_fun (callable): Function that constructs the ansatz (trial and test) functions.
    ref_int_coor (jnp.ndarray): Reference integration coordinates for the isoparametric element.
    ref_int_weights (jnp.ndarray): Reference integration weights for the isoparametric element.
    tangent_contributions (bool): If True, compute the tangent contributions; otherwise, return zero tangent.
    initial_config (bool, optional): If True, use the initial configuration for isoparametric mapping. Default is True.

  Returns:
    callable:
      A function that evaluates the residual and tangent matrix for the isoparametric surface element.

  Notes:
    - Currently, the weak form function must not contain derivatives wrt. the primary fields.
    - The returned function can be used to evaluate the residual and tangent matrix for the element.
    - The mode parameter in the returned function can be 'residual' or 'tangent'.

  The returned function has the following parameters:
    - fI (jnp.ndarray): Element nodal values.
    - xI (jnp.ndarray): Element nodal coordinates.
    - elem_number (int): Element number.
    - settings (dict): Settings for the computation.
    - static_settings (dict): Static settings for the computation.
    - mode (str): Specifies whether to compute the 'residual' or 'tangent'.
  """

  def user_element(fI, xI, elem_number, settings, static_settings, mode, set):
    n_dofs = fI.flatten().shape[0]
    n_dim = xI.shape[-1]

    # One Gauß point contribution
    def residual_and_tangent(x_int, w_int):

      def integrated_weak_form(test_fI, trial_fI):
        # Define trial and test function ansatz
        trial_ansatz = lambda x: ansatz_fun(x, xI, trial_fI, settings, False, n_dim-1)
        test_ansatz = lambda x: ansatz_fun(x, xI, test_fI, settings, False, n_dim-1)

        # Evaluate weak form
        discrete_weak_form = weak_form_fun(x_int, trial_ansatz, test_ansatz, settings, static_settings, elem_number, set)

        # Compute the Jacobian
        if initial_config:
          mapping = lambda x: ansatz_fun(x, xI, xI, settings, False, n_dim-1)
        else:
          # Take first n_dim fields as displacements
          uI = fI[:, :xI.shape[-1]]
          mapping = lambda x: ansatz_fun(x, xI + uI, xI + uI, settings, overwrite_diff=False)
        jacobian = jax.jacrev(mapping)(x_int)

        match n_dim:
          case 2:
            G_1 = jnp.concatenate([jacobian, jnp.asarray([0])])
            E_3 = jnp.asarray([0, 0, 1])
            scaling = jnp.linalg.norm(jnp.cross(G_1, E_3))
          case 3:
            G_1 = jacobian[:,0]
            G_2 = jacobian[:,1]
            scaling = jnp.linalg.norm(jnp.cross(G_1, G_2))
          case _:
            assert False, "Surface element supports currently only lines in 2d and 2d-surfaces in 3d."

        return w_int * scaling * discrete_weak_form

      # Residual
      residual_fun = jax.jacrev(integrated_weak_form, argnums=0)
      residual = residual_fun(fI, fI).flatten()

      # Tangent
      if tangent_contributions:
        tangent_fun = jax.jacfwd(residual_fun, argnums=1)
        tangent = tangent_fun(fI, fI).reshape(n_dofs,n_dofs)
      else:
        tangent = jnp.zeros((n_dofs, n_dofs), dtype=jnp.float_)

      return (residual, tangent)

    # Sum over Gauß points
    (residual_contrib, tangent_contrib) = jax.vmap(residual_and_tangent, (0, 0), (0, 0))(ref_int_coor, ref_int_weights)
    residual = residual_contrib.sum(axis=0)
    tangent = tangent_contrib.sum(axis=0)

    if mode == 'residual':
      return residual
    elif mode =='tangent':
      return tangent
    else:
      assert False, 'User elem mode has to be \'residual\' or \'tangent\'.'

  return user_element


### Time integration procedures
def forward_backward_euler_weak(inertia_coeff_fun):
  """
  Constructs a weak form time discretization function using the Forward or Backward Euler method.

  For an examplary use, see the examples with explicit and implicit time integration.

  Parameters:
    inertia_coeff_fun (callable): Function to compute the inertia coefficient given spatial coordinates and settings.

  Returns:
    callable:
      A function that evaluates the weak form time discretization for the PDE.

  The returned function has the following parameters:
    - x (jnp.ndarray): Spatial coordinates at the integration point.
    - ansatz (callable): Ansatz function representing the primary field at time step n+1.
    - test_ansatz (callable): Test ansatz function representing the test function.
    - settings (dict): Settings for the computation, including:
      - 'dofs n': Degrees of freedom at the previous time step.
      - 'connectivity': Connectivity information for the elements.
      - 'time increment': Time increment delta_t.
    - static_settings (dict): Static settings for the computation.
    - int_point_number (int): Integration point number.
    - set: Number of domain

  Notes:
    - The primary field is evaluated at both time steps n and n+1.
    - The time derivative is computed using the Forward or Backward Euler method, depending on the use case.
    - The inertia coefficient is used to scale the time derivative term.
  """
  def time_discretization_fun(x, ansatz, test_ansatz, settings, static_settings, int_point_number, set):
    # Test function
    test_function = test_ansatz(x)

    # Primary field evaluated at n and n+1
    dofs_n = settings['dofs n']
    neighbor_list = jnp.asarray(static_settings['connectivity'][set])[int_point_number]
    local_dofs_n = dofs_n[neighbor_list]
    ansatz_n = lambda x: solution_structures.solution_structure(x, int_point_number, local_dofs_n, settings, static_settings, set)
    field_n = ansatz_n(x)
    field = ansatz(x)

    # Forward/Backward Euler
    delta_t = settings['time increment']
    time_derivative = (field - field_n) / delta_t

    # E.g. heat capacity or whatever is multiplied with first order time derivative
    inertia_coeff = inertia_coeff_fun(x, settings)
    return - inertia_coeff * jnp.dot(test_function, time_derivative)
  return time_discretization_fun

def central_differences(density_fun, damping):
  """
  Constructs weak form contributions for acceleration (and optional damping) term based on central differences.

  Parameters:
    density_fun (callable): Function to compute the density coefficient given spatial coordinates and settings.
    damping (bool): Whether to turn on damping. Damping requires settings['damping coefficient'] to be set.

  Returns:
    callable:
      A function that evaluates the weak form contributions.

  The returned function has the following parameters:
    - x (jnp.ndarray): Spatial coordinates at the integration point.
    - ansatz (callable): Ansatz function representing the primary field at time step n+1.
    - test_ansatz (callable): Test ansatz function representing the test function.
    - settings (dict): Settings for the computation, including:
      - 'dofs n': Degrees of freedom at the previous time step.
      - 'connectivity': Connectivity information for the elements.
      - 'time increment': Time increment delta_t.
    - static_settings (dict): Static settings for the computation.
    - int_point_number (int): Integration point number.
    - set: Number of domain

  Notes:
    - Requires settings['dofs n], settings['dofs n-1'] and settings['time increment'].
  """
  def weak_form_fun(x, ansatz, test_ansatz, settings, static_settings, int_point_number, set):

    # Test function
    test_function = test_ansatz(x)

    # Primary field evaluated at n-1, n and n+1
    dofs_n = settings['dofs n']
    dofs_n_min_1 = settings['dofs n-1']
    neighbor_list = jnp.asarray(static_settings['connectivity'][set])[int_point_number]
    ansatz_n = lambda x: solution_structures.solution_structure(x, int_point_number, dofs_n[neighbor_list], settings, static_settings, set)
    ansatz_n_min_1 = lambda x: solution_structures.solution_structure(x, int_point_number, dofs_n_min_1[neighbor_list], settings, static_settings, set)
    u_n = ansatz_n(x)
    u_n_min_1 = ansatz_n_min_1(x)
    u_n_plus_1 = ansatz(x)


    # Central differences
    delta_t = settings['time increment']
    a_n = (u_n_plus_1 - 2 * u_n - u_n_min_1) / (delta_t**2)
    v_n = (u_n_plus_1 - u_n_min_1) / (2 * delta_t)

    # E.g. heat capacity or whatever is multiplied with first order time derivative
    density = density_fun(x, settings)
    if damping:
      damping_coeff = settings['damping coefficient']
      return density * jnp.dot(test_function, a_n + damping_coeff * v_n)
    else:
      return density * jnp.dot(test_function, a_n)
  return weak_form_fun



# plotter.py
# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
This module provides some convenience functions for interactive 3D plotting and exporting visualizations using PyVista and Plotly.
"""


import os
import jax
import jax.numpy as jnp
import numpy as np


def pv_plot(x_vis, data, export_vtk=False, show=True, groupname='group', filename='frame', fieldname='data', scale_range=1.0, data_on_z=True):
  """
  Plots and optionally exports 3D data using PyVista.

  Parameters:
    - x_vis (np.ndarray): Array of coordinates for visualization points.
    - data (np.ndarray): Data to be visualized at the points.
    - export_vtk (bool, optional): Whether to export the plot to a VTK file. Default is False.
    - show (bool, optional): Whether to show the plot. Default is True.
    - groupname (str, optional): Name of the group folder for VTK export. Default is 'group'.
    - filename (str, optional): Name of the file for VTK export. Default is 'frame'.
    - fieldname (str, optional): Name of the data field in the VTK file. Default is 'data'.
    - scale_range (float, optional): Scale range for data visualization. Default is 1.0.
    - data_on_z (bool, optional): Whether to plot data on the z-axis. Default is True.

  Notes:
    - Only supports 2d and 3d
  """
  import pyvista as pv
  pv.global_theme.cmap = 'jet'

  if x_vis.shape[1] not in [2, 3]:
      raise AssertionError("Currently not implemented for dimensions other than 2 or 3.")

  if x_vis.shape[1] == 2:
      x_range = jnp.nanmax(x_vis.flatten()) - jnp.nanmin(x_vis.flatten())
      data_range = jnp.nanmax(data) - jnp.nanmin(data)
      scaled_data = scale_range * (x_range / data_range) * data if data_range > 1e-8 else data
      positions = np.column_stack((x_vis[:, 0], x_vis[:, 1], scaled_data if data_on_z else np.zeros_like(scaled_data)))
  else:
      positions = np.asarray(x_vis)

  points = pv.PolyData(positions)
  points.point_data[fieldname] = np.asarray(data)

  if show:
      points.plot(point_size=3, style='points')

  if export_vtk:
      os.makedirs(groupname, exist_ok=True)
      points.save(os.path.join(groupname, f'{filename}.vtp'))

def pv_function_plot(x_vis, fun, scale_range=1.0, data_on_z=True, isosurf=False, n_surfaces=20):
  """
  Convenience function for visualizing/debugging smooth distance functions.

  Parameters:
    - x_vis (np.ndarray): Array containing the coordinates of nodes for visualization.
    - fun (function): Function to be visualized depending on x.
    - scale_range (float, optional): Scale range for data visualization. Default is 1.0.
    - data_on_z (bool, optional): Whether to plot data on the z-axis. Default is True.
    - isosurf (bool, optional): Whether to plot isosurfaces for structured 3D grids. Default is False.
    - n_surfaces (int, optional): Number of isosurfaces to plot. Default is 20.
  """
  data = jax.jit(jax.vmap(fun, (0), 0))(x_vis)
  if isosurf:
      isosurface(x_vis, data, n_surfaces)
  else:
      pv_plot(x_vis, data, export_vtk=False, show=True, scale_range=scale_range, data_on_z=data_on_z)

def isosurface(x_vis, data, n_surfaces=20):
  """
  Plots isosurfaces for structured 3D data using Plotly.

  Parameters:
    - x_vis (np.ndarray): Array of coordinates for visualization points.
    - data (np.ndarray): Data to be visualized at the points.
    - n_surfaces (int, optional): Number of isosurfaces to plot. Default is 20.
  """
  import plotly.graph_objects as go

  fig = go.Figure(data=go.Isosurface(
      x=x_vis[:, 0],
      y=x_vis[:, 1],
      z=x_vis[:, 2],
      value=data,
      opacity=0.5,
      isomin=jnp.nanmin(data).item(),
      isomax=jnp.nanmax(data).item(),
      colorscale='Rainbow',
      surface_count=n_surfaces,
      caps=dict(x_show=False, y_show=False)
  ))
  fig.show()




# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

import time
import sys

import jax
from jax import config
import jax.numpy as jnp
import flax
import pygmsh
import meshio

from autopdex import seeder, geometry, solver, utility, models, spaces

config.update("jax_enable_x64", True)
config.update("jax_compilation_cache_dir", './cache')


### Cook's membrane, see e.g. https://link.springer.com/article/10.1007/s00466-017-1405-4
# Refinement with isoparametric quadrilaterls of order p

# p refinement
post_values = []
for order in range(1,10):

  ### Definition of geometry and boundary conditions
  pts = [[0.,0.], [48.,44.], [48.,60.], [0.,44.]]
  with pygmsh.occ.Geometry() as geom:
      region = geom.add_polygon(pts, mesh_size=10.0)
      geom.set_recombined_surfaces([region.surface])
      mesh = geom.generate_mesh(order=order)
      print("Mesh generation finished.")

  # Import region mesh
  n_dim = 2
  x_nodes = jnp.asarray(mesh.points[:,:n_dim])
  n_nodes = x_nodes.shape[0]
  elements = jnp.asarray([ v for k, v in mesh.cells_dict.items() if 'quad' in k])[0]
  surface_elements = jnp.asarray([ v for k, v in mesh.cells_dict.items() if 'line' in k])[0]
  print("Number of elements: ", elements.shape[0])
  print("Number of unknowns: ", n_nodes * n_dim)

  # Selection of nodes for boundary condition
  dirichlet_nodes = geometry.in_planes(x_nodes, pts[0], [1.,0.])
  dirichlet_dofs = utility.dof_select(dirichlet_nodes, jnp.asarray([True, True]))
  dirichlet_conditions = jnp.zeros_like(dirichlet_dofs, dtype=jnp.float_)

  # Import surface mesh for inhomogeneous Neumann conditions
  neumann_selection = geometry.select_elements_in_plane(x_nodes, surface_elements, pts[1], [1.,0.])
  neumann_elements= surface_elements[neumann_selection]

  # Definition of pde
  lam = 100.
  mu = 40.
  Em = mu * (3 * lam + 2 * mu) / (lam + mu)
  nu = lam / (2 * (lam + mu))
  youngs_mod_fun = lambda x: Em
  poisson_ratio_fun = lambda x: nu
  weak_form_fun_1 = models.linear_elasticity_weak(youngs_mod_fun, poisson_ratio_fun, 'plain strain') # small deformations
  # weak_form_fun_1 = models.hyperelastic_steady_state_weak(models.neo_hooke, youngs_mod_fun, poisson_ratio_fun, 'plain strain') # large deformations
  user_elem_1 = models.isoparametric_domain_element_galerkin(weak_form_fun_1, 
                                                          spaces.fem_iso_line_quad_brick,
                                                          *seeder.gauss_legendre_nd(dimension = 2, order = 2 * order))

  # Tractions
  q_0 = 4.0e+0
  # traction_fun = lambda x: jnp.asarray([0., q_0])
  traction_fun = lambda x, settings: jnp.asarray([0., settings['load multiplier']])
  weak_form_fun_2 = models.neumann_weak(traction_fun)
  user_elem_2 = models.isoparametric_surface_element_galerkin(weak_form_fun_2, 
                                                          spaces.fem_iso_line_quad_brick,
                                                          *seeder.gauss_legendre_nd(dimension = 1, order = 2 * order),
                                                          tangent_contributions=False)

  ### Settings
  n_fields = 2
  dofs0 = jnp.zeros((n_nodes, n_fields))
  static_settings = flax.core.FrozenDict({
    'number of fields': (n_fields, n_fields),
    'assembling mode': ('user element', 'user element'),
    'solution structure': ('nodal imposition', 'nodal imposition'),
    'model': (user_elem_1, user_elem_2),
    'solver type': 'newton',
    'solver backend': 'pardiso',
    'solver': 'lu',
    'verbose': 0,
    'dirichlet dofs': utility.jnp_to_tuple(dirichlet_dofs),
    'connectivity': (utility.jnp_to_tuple(elements),utility.jnp_to_tuple(neumann_elements)),
  })

  settings = {
    'load multiplier': q_0,
    'node coordinates': x_nodes,
    'dirichlet conditions': dirichlet_conditions,
  }

  ### Manual load stepping
  start = time.time()
  dofs = dofs0
  multiplier = 0.0
  n_step = 2
  for i in range(n_step):
    # Update boundary conditions
    print('Start load step ', i+1)
    multiplier = multiplier + 1 / n_step
    settings['load multiplier'] = multiplier * q_0

    # Call newton solver
    dofs, _ = solver.solver(dofs, settings, static_settings)


  # Paraview postprocessing
  points = mesh.points
  cells = jnp.asarray([ v for k, v in mesh.cells_dict.items() if 'quad' in k])[0]
  mesh = meshio.Mesh(
      points,
      {'quad': cells[:,:4]},
      point_data={
          "u": dofs,
      },
  )
  mesh.write("cooks_membrane.vtk")

  # Displacement at upper right edge
  idx = geometry.select_point(x_nodes, pts[2])
  v_ul = dofs[idx, 1].item()
  post_values.append([n_nodes * n_dim, v_ul]) # number of dofs over edge displacement in y-direction

print("Vertical displacements at upper right edge for each p-order:\n", post_values)



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

def test_example_lid_driven_cavity():
    import sys

    import jax
    from jax import config
    import jax.numpy as jnp
    import flax
    import pyvista as pv
    import numpy as np

    from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models, spaces, assembler

    config.update("jax_enable_x64", True)

    ### Definition of geometry and boundary conditions
    # Positive smooth distance functions that select the specific boundary segments
    pts = [[0.,0.], [0.1,0.], [0.1,0.1], [0.,0.1]]
    psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, pts[0], pts[1])
    psdf_right = lambda x: geometry.psdf_trimmed_line(x, pts[1], pts[2])
    psdf_top = lambda x: geometry.psdf_trimmed_line(x, pts[2], pts[3])
    psdf_left = lambda x: geometry.psdf_trimmed_line(x, pts[3], pts[0])
    psdf_rect = lambda x: geometry.psdf_parallelogram(x, pts[0], pts[3], pts[1])
    psdfs = [psdf_bottom, psdf_right, psdf_top, psdf_left]

    # Positive smooth distance funcion of full domain
    psdf = lambda x: jnp.asarray([psdf_rect(x), psdf_rect(x)])

    # Field selectors for boundary conditions
    selection_dirichlet_1 = lambda x: jnp.asarray([0.,1.,0.])
    selection_dirichlet_2 = lambda x: jnp.asarray([0.,0.,1.])

    # Assignment of boundary conditions (same ordering as in psdfs)
    homogeneous = lambda x: 0.
    one = lambda x: 1.
    bc_1 = [homogeneous, homogeneous, one, homogeneous]

    ### Transfinite interpolation of boundary segments
    b_coeff = lambda x: jnp.asarray([selection_dirichlet_1(x),
                                    selection_dirichlet_2(x)])
    b_cond = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, bc_1),
                                    homogeneous(x)])


    ### Initialization and generation of nodes and integration points
    # Regular node placement
    nod_spacing = 0.02
    (x_nodes, n_nodes, region_size) = seeder.regular_in_psdf(psdf_rect, pts[0], pts[2], nod_spacing)

    # Gauss integration on background grid
    int_spacing = nod_spacing
    (x_int, w_int, n_int) = seeder.gauss_points_in_psdf(psdf_rect, pts[0], pts[2], int_spacing, order=2)

    # Support radius for moving least square ansatz
    support_radius = 3.01 * nod_spacing

    # Definition of pde
    pde = models.navier_stokes_incompressible_steady(dynamic_viscosity=0.01)

    ### Neighborhood search
    (num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)


    ### Setting
    n_fields = 3
    dofs0 = jnp.zeros((n_nodes, n_fields))
    static_settings = flax.core.FrozenDict({
    'solution space': ('mls',),
    'shape function mode': 'compiled',
    'number of fields': (n_fields,),
    'assembling mode': ('sparse',),
    'maximal number of neighbors': (max_neighbors,),
    'order of basis functions': (2,),
    'weight function type': ('gaussian',),

    # Note the inaccuracy in the pressure field with least square variational scheme
    # 'variational scheme': ('least square pde loss',),
    'variational scheme': ('strong form galerkin',),

    'model': (pde,),
    'solution structure': ('second order set',),
    'boundary coefficients': (b_coeff,),
    'boundary conditions': (b_cond,),
    'psdf': (psdf,),
    'solver type': 'newton',
    'solver backend': 'scipy',
    'solver': 'lapack',
    'verbose': -1,
    'connectivity': (utility.jnp_to_tuple(neighbor_list),),
    })
    settings = {
    'beta': (3,),
    'node coordinates': x_nodes,
    'integration coordinates': (x_int,),
    'integration weights': (w_int,),
    'support radius': (support_radius,),
    }


    ### Optional: Precompute shape functions and pass them to e.g. static_settings...
    settings = solution_structures.precompile(dofs0, settings, static_settings)

    tangent = assembler.assemble_tangent(dofs0, settings, static_settings)
    assert jnp.isclose(tangent.data.flatten().sum(), 0.008454379736301122)



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
This example addresses Cook's membrane problem where the Young's modulus 
and Poisson's ratio follow a multivariate normal distribution. The goal 
is to compute the expectation and standard deviation of the displacement 
field. Two approaches are compared: the Monte Carlo method and an approach
based on a Taylor series expansion. For the Taylor series expansion, 
the derivatives of the displacement degrees of freedom with respect to 
Young's modulus and Poisson's ratio are required. These sensitivities 
are calculated using automatic implicit differentiation based on the 
implicit function theorem.
"""


if __name__ == "__main__":
    import time
    import sys

    import numpy as np
    import meshio
    import pygmsh
    import jax
    from jax import config
    import jax.numpy as jnp
    import flax

    from autopdex import assembler, seeder, geometry, solver, utility, models, spaces, implicit_diff

    config.update("jax_enable_x64", True)
    config.update("jax_compilation_cache_dir", './cache')

    # Order of ansatz
    order = 2

    ### Definition of geometry and boundary conditions
    pts = [[0.,0.], [48.,44.], [48.,60.], [0.,44.]]
    with pygmsh.occ.Geometry() as geom:
        # region = geom.add_polygon(pts, mesh_size=10.0)
        region = geom.add_polygon(pts, mesh_size=1.0)
        geom.set_recombined_surfaces([region.surface])
        mesh = geom.generate_mesh(order=order)
        print("Mesh generation finished.")

    # Import mesh
    n_dim = 2
    x_nodes = jnp.asarray(mesh.points[:,:n_dim])
    n_nodes = x_nodes.shape[0]
    print("Number of nodes: ", n_nodes)
    elements = jnp.asarray([ v for k, v in mesh.cells_dict.items() if 'quad' in k])[0]
    surface_elements = jnp.asarray([ v for k, v in mesh.cells_dict.items() if 'line' in k])[0]

    # Selection of nodes for boundary condition
    dirichlet_nodes = geometry.in_planes(x_nodes, pts[0], [1.,0.])
    dirichlet_dofs = utility.dof_select(dirichlet_nodes, jnp.asarray([True, True]))
    dirichlet_conditions = jnp.zeros_like(dirichlet_dofs, dtype=jnp.float_)

    # Import surface mesh for inhomogeneous Neumann conditions
    neumann_selection = geometry.select_elements_in_plane(x_nodes, surface_elements, pts[1], [1.,0.])
    neumann_elements= surface_elements[neumann_selection]


    ### Built-in autopdex element
    weak_form_fun_1 = models.hyperelastic_steady_state_weak(
    models.neo_hooke, 
    lambda x, settings: settings['youngs modulus'], # Youngs modulus loaded from settings
    lambda x, settings: settings['poisson ratio'], 
    'plain strain')
    user_elem_1 = models.isoparametric_domain_element_galerkin(
    weak_form_fun_1, 
    spaces.fem_iso_line_quad_brick,
    *seeder.gauss_legendre_nd(dimension = 2, order = 2 * order))

    # Tractions
    q_0 = 2.0e+1
    traction_fun = lambda x, settings: jnp.asarray([0., settings['load multiplier']])
    weak_form_fun_2 = models.neumann_weak(traction_fun)
    user_elem_2 = models.isoparametric_surface_element_galerkin(
    weak_form_fun_2, 
    spaces.fem_iso_line_quad_brick,
    *seeder.gauss_legendre_nd(dimension = 1, order = 2 * order),
    tangent_contributions=False)

    # Adaptive load stepping
    def multiplier_settings(settings, multiplier):
        settings['load multiplier'] = multiplier * q_0
        return settings

    ### Settings
    n_fields = 2
    dofs_0 = jnp.zeros((n_nodes, n_fields))
    static_settings = flax.core.FrozenDict({
    'number of fields': (n_fields, n_fields),
    'assembling mode': ('user element', 'user element'),
    'solution structure': ('nodal imposition', 'nodal imposition'),
    'model': (user_elem_1, user_elem_2),
    'solver type': 'newton',
    'solver backend': 'pardiso',
    'solver': 'lu',
    'verbose': 0,
    'dirichlet dofs': utility.jnp_to_tuple(dirichlet_dofs),
    'connectivity': (utility.jnp_to_tuple(elements), utility.jnp_to_tuple(neumann_elements)),
    })
    settings = {
    'load multiplier': q_0,
    'node coordinates': x_nodes,
    'dirichlet conditions': dirichlet_conditions,
    }



    ### Switch: Monte Carlo/ Taylor series
    monte_carlo = False
    key = jax.random.PRNGKey(123)
    if monte_carlo:
        n_monte_carlo = 1000
    else:
        n_monte_carlo = 10000

    # Generate samples
    means = jnp.asarray([100., 0.3])
    covs = jnp.asarray([[10., 0.0001], [0.0001, 0.001]])
    samples = jax.random.multivariate_normal(key, means, covs, (n_monte_carlo,))
    youngs_mod_mc = samples[:,0]
    nu_mc = samples[:,1]

    ### Estimate uncertainties with Monte Carlo
    start = time.time()
    if monte_carlo:
        key = jax.random.PRNGKey(123)

        # Generate samples
        means = jnp.asarray([100., 0.3])
        covs = jnp.asarray([[10., 0.0001], [0.0001, 0.001]])
        samples = jax.random.multivariate_normal(key, means, covs, (n_monte_carlo,))
        youngs_mod_mc = samples[:,0]
        nu_mc = samples[:,1]

        # Check the stochasticity of the material parameters
        mean_Em_test = youngs_mod_mc.sum()/n_monte_carlo
        variance_Em_test = jnp.sum((youngs_mod_mc - mean_Em_test)**2) / n_monte_carlo
        print("Expectation value of youngs modulus: ", mean_Em_test)
        print("Variance of youngs modulus: ", variance_Em_test)

        mean_nu_test = nu_mc.sum()/n_monte_carlo
        variance_nu_test = jnp.sum((nu_mc - mean_nu_test)**2) / n_monte_carlo
        print("Expectation value of Poisson ratio: ", mean_nu_test)
        print("Variance of Poisson ratio: ", variance_nu_test)

        @jax.jit
        def compute_dofs(dofs, settings, Em, nu):
            settings['youngs modulus'] = Em
            settings['poisson ratio'] = nu
            return solver.adaptive_load_stepping(dofs, 
                                                 settings, 
                                                 static_settings, 
                                                 multiplier_settings, 
                                                 False, 
                                                 'reverse',
                                                 min_increment=0.001, 
                                                 newton_tol=1e-8)[0]

        # Batched forward analysis and uncertainty estimation via Monte Carlo
        dofs_mc = jnp.asarray([compute_dofs(dofs_0, settings, youngs_mod_mc[i], nu_mc[i]) for i in range(youngs_mod_mc.shape[0])])
        mean_dofs = jnp.sum(dofs_mc, axis=0) / n_monte_carlo
        variance_dofs = jnp.sum((dofs_mc - mean_dofs)**2, axis=0) / n_monte_carlo
        std_dofs = jnp.sqrt(variance_dofs)

        skewness_dofs = jnp.sum((dofs_mc - mean_dofs)**3, axis=0) / n_monte_carlo / std_dofs**3
        kurtosis_dofs = jnp.sum((dofs_mc - mean_dofs)**4, axis=0) / n_monte_carlo / std_dofs**4

        print("Time needed for Monte Carlo analysis: ", time.time() - start)

        ## Postprocessing
        points = mesh.points
        cells = jnp.asarray([ v for k, v in mesh.cells_dict.items() if 'quad' in k])[0]
        mesh = meshio.Mesh(
            points,
            {'quad': cells[:,:4]},
            point_data={
                "expectation(u)": mean_dofs,
                "variance(u)": variance_dofs,
                "std_deviation(u)": std_dofs,
                "skewness(u)": skewness_dofs,
                "kurtosis(u)": kurtosis_dofs,
            },
        )
        mesh.write("uncertainties_monte_carlo.vtk")

    ### Estimate uncertainties via Taylor series with implicit differentiation
    else:
        youngs_mod_mean = means[0]
        nu_mean = means[1]

        def compute_dofs(dofs, settings, Em, nu):
            settings['youngs modulus'] = Em
            settings['poisson ratio'] = nu
            return solver.adaptive_load_stepping(dofs, settings, static_settings, multiplier_settings, False, 'forward', newton_tol=1e-8)[0]


        ## Sensitivity analysis
        # # Version 1 using jacfwd_upto_n_scalar_args
        # forward_and_sensitivity = lambda Em, nu: compute_dofs(dofs_0, settings, Em, nu)
        # result = utility.jacfwd_upto_n_scalar_args(forward_and_sensitivity, (youngs_mod_mean, nu_mean), 1, (0, 1))
        # dofs, du_dEm, du_dnu = result[0,0], result[0,1], result[1,1]

        # # Higher order sensitivities (needs longer compile time...)
        # forward_and_sensitivity = lambda Em, nu: compute_dofs(dofs_0, settings, Em, nu)
        # result = utility.jacfwd_upto_n_scalar_args(forward_and_sensitivity, (youngs_mod_mean, nu_mean), 3, (0, 1))
        # dofs, du_dEm, du_dnu, d2u_dEm2, d2u_dnu2, d3u_dEm3, d3u_dnu3 = result[0,0], result[1,0], result[1,1], result[2,0], result[2,1], result[3,0], result[3,1]

        # # Version 2 using jacfwd_upto_n_one_vector_arg
        # forward_and_sensitivity = lambda v: compute_dofs(dofs_0, settings, v[0], v[1])
        # result = utility.jacfwd_upto_n_one_vector_arg(forward_and_sensitivity, jnp.asarray([youngs_mod_mean, nu_mean]), 1)
        # dofs, du_dEm, du_dnu = result[0], result[1][0], result[1][1]

        # Second order mixed sensitivities
        forward_and_sensitivity = lambda v: compute_dofs(dofs_0, settings, v[0], v[1])
        result = utility.jacfwd_upto_n_one_vector_arg(forward_and_sensitivity, jnp.asarray([youngs_mod_mean, nu_mean]), 2)
        dofs, du_dEm, du_dnu, d2u_dEm2, d2u_dEmdnu, d2u_dnudEm, d2u_dnu2 = result[0], result[1][0], result[1][1], result[2][0][0], result[2][0][1], result[2][1][0], result[2][1][1]

        # # Third order mixed sensitivities (needs longer compile time...)
        # forward_and_sensitivity = lambda v: compute_dofs(dofs_0, settings, v[0], v[1])
        # result = utility.jacfwd_upto_n_one_vector_arg(forward_and_sensitivity, jnp.asarray([youngs_mod_mean, nu_mean]), 3)
        # dofs, du_dEm, du_dnu, d2u_dEm2, d2u_dEmdnu, d2u_dnudEm, d2u_dnu2, d3u_dnu3, d3u_dnu2dEm = result[0], result[1][0], result[1][1], result[2][0][0], result[2][0][1], result[2][1][0], result[2][1][1], result[3][1][1][1], result[3][1][1][0]

        print("Sensitivity analysis finished. Starting uncertainty estimation...")

        ## First order Taylor series expansion
        # # Estimate expectation of displacement field via first-order Taylor expansions
        # def dofs_i(dofs, du_dEm, du_dnu, youngs_mod_i, nu_i):
        #     return dofs + du_dEm * (youngs_mod_i - youngs_mod_mean) + du_dnu * (nu_i - nu_mean)
        # mean_dofs = jnp.sum(jax.jit(jax.vmap(dofs_i, (None, None, None, 0, 0), 0))(dofs, du_dEm, du_dnu, youngs_mod_mc, nu_mc), axis=0) / n_monte_carlo

        # # Estimate variance, standard deviation, skewness, and kurtosis of displacement field via first-order Taylor expansions with monte carlo integration
        # def moments_i(dofs, du_dEm, du_dnu, youngs_mod_i, nu_i):
        #     delta = du_dEm * (youngs_mod_i - youngs_mod_mean) + du_dnu * (nu_i - nu_mean) + dofs - mean_dofs
        #     var = delta**2
        #     skew = delta**3
        #     kurt = delta**4
        #     return var, skew, kurt

        # # Calculate moments
        # var_skew_kurt = jax.jit(jax.vmap(moments_i, (None, None, None, 0, 0), 0))(dofs, du_dEm, du_dnu, youngs_mod_mc, nu_mc)

        # # Variance and standard deviation
        # variance_dofs = jnp.sum(var_skew_kurt[0], axis=0) / n_monte_carlo
        # std_dofs = jnp.sqrt(variance_dofs)

        # # Skewness: normalized third central moment
        # skewness_dofs = jnp.sum(var_skew_kurt[1], axis=0) / n_monte_carlo / (std_dofs**3)
        
        # # Kurtosis: normalized fourth central moment
        # kurtosis_dofs = jnp.sum(var_skew_kurt[2], axis=0) / n_monte_carlo / (std_dofs**4)

        ## Second order Taylor series expansion
        @jax.jit
        def compute_mean_var_skew_kurst(dofs, du_dEm, du_dnu, d2u_dEm2, d2u_dnu2, d2u_dEmdnu, youngs_mod_mc, nu_mc):
            # Estimate expectation of displacement field via second-order Taylor expansions
            def dofs_i(dofs, du_dEm, du_dnu, d2u_dEm2, d2u_dnu2, d2u_dEmdnu, youngs_mod_i, nu_i):
                delta_Em = youngs_mod_i - youngs_mod_mean
                delta_nu = nu_i - nu_mean
                linear_term = du_dEm * delta_Em + du_dnu * delta_nu
                quadratic_term = 0.5 * (d2u_dEm2 * delta_Em**2 + d2u_dnu2 * delta_nu**2 + 2 * d2u_dEmdnu * delta_Em * delta_nu)
                return dofs + linear_term + quadratic_term
            mean_dofs = jnp.sum(jax.jit(jax.vmap(dofs_i, (None, None, None, None, None, None, 0, 0), 0))(
                dofs, du_dEm, du_dnu, d2u_dEm2, d2u_dnu2, d2u_dEmdnu, youngs_mod_mc, nu_mc), axis=0) / n_monte_carlo

            # Estimate variance, standard deviation, skewness, and kurtosis of displacement field via second-order Taylor expansions with monte carlo integration
            def moments_i(dofs, du_dEm, du_dnu, d2u_dEm2, d2u_dnu2, d2u_dEmdnu, youngs_mod_i, nu_i):
                delta_Em = youngs_mod_i - youngs_mod_mean
                delta_nu = nu_i - nu_mean
                linear_term = du_dEm * delta_Em + du_dnu * delta_nu
                quadratic_term = 0.5 * (d2u_dEm2 * delta_Em**2 + d2u_dnu2 * delta_nu**2 + 2 * d2u_dEmdnu * delta_Em * delta_nu)
                delta = linear_term + quadratic_term + dofs - mean_dofs
                var = delta**2
                skew = delta**3
                kurt = delta**4
                return var, skew, kurt

            # Calculate moments with second-order Taylor expansion
            var_skew_kurt = jax.jit(jax.vmap(moments_i, (None, None, None, None, None, None, 0, 0), 0))(
                dofs, du_dEm, du_dnu, d2u_dEm2, d2u_dnu2, d2u_dEmdnu, youngs_mod_mc, nu_mc)

            # Variance and standard deviation
            variance_dofs = jnp.sum(var_skew_kurt[0], axis=0) / n_monte_carlo
            std_dofs = jnp.sqrt(variance_dofs)

            # Skewness: normalized third central moment
            skewness_dofs = jnp.sum(var_skew_kurt[1], axis=0) / n_monte_carlo / (std_dofs**3)

            # Kurtosis: normalized fourth central moment
            kurtosis_dofs = jnp.sum(var_skew_kurt[2], axis=0) / n_monte_carlo / (std_dofs**4)
            return mean_dofs, variance_dofs, std_dofs, skewness_dofs, kurtosis_dofs
        mean_dofs, variance_dofs, std_dofs, skewness_dofs, kurtosis_dofs = compute_mean_var_skew_kurst(dofs, du_dEm, du_dnu, d2u_dEm2, d2u_dnu2, d2u_dEmdnu, youngs_mod_mc, nu_mc)

        ## Postprocessing
        print("Time needed for Taylor series/ implicit differentiation analysis: ", time.time() - start)
        points = mesh.points
        cells = jnp.asarray([ v for k, v in mesh.cells_dict.items() if 'quad' in k])[0]
        mesh = meshio.Mesh(
            points,
            {'quad': cells[:,:4]},
            point_data={
                "u": dofs,
                "expectation(u)": mean_dofs,
                "variance(u)": variance_dofs,
                "std_deviation(u)": std_dofs,
                "skewness(u)": skewness_dofs,
                "kurtosis(u)": kurtosis_dofs,
                "du/dEm": du_dEm,
                "du/dnu": du_dnu,

                # "d2u/dnu2": d2u_dnu2,
                # "d2u/dEm2": d2u_dEm2,
                # "d3u/dnu3": d3u_dnu3,
                # "d3u/dEm3": d3u_dEm3,

                "d2u/dnu2": d2u_dnu2,
                "d2u/dEmdnu": d2u_dEmdnu,
                "d2u/dnudEm": d2u_dnudEm,
                "d2u/dEm2": d2u_dEm2,
                
                # "d3u/dnu3": d3u_dnu3,
                # "d3u/dnu2dEm": d3u_dnu2dEm,
            },
        )
        mesh.write("uncertainties_taylor.vtk")



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

import sys

import jax
from jax import config
import jax.numpy as jnp
import numpy as np
import flax
import pygmsh
import meshio

from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models, assembler

config.update("jax_enable_x64", True)
config.update("jax_compilation_cache_dir", './cache')


### Cook's membrane problem with non-homogeneous Dirichlet conditions nodally imposed

### Definition of geometry and boundary conditions
pts = [[0.,0.], [48.,44.], [48.,60.], [0.,44.]]
with pygmsh.geo.Geometry() as geom:
    geom.add_polygon(pts,mesh_size=2.0)
    mesh = geom.generate_mesh(order=1)

# Mesh debugging
mesh.write("mesh.vtk")

# Import region mesh
n_dim = 2
x_nodes = jnp.asarray(mesh.points[:,:n_dim])
n_nodes = x_nodes.shape[0]
elements = jnp.asarray(mesh.cells_dict['triangle'])
surface_elements = jnp.asarray(mesh.cells_dict['line'])

# Selection of nodes for boundary condition
dirichlet_left = geometry.in_planes(x_nodes, pts[0], [1.,0.])
dirichlet_dofs_left = utility.dof_select(dirichlet_left, jnp.asarray([True, True]))
dirichlet_conditions = jnp.zeros_like(dirichlet_dofs_left, dtype=jnp.float_)

# Assignment of boundary conditions at right edge
v_bar = lambda x: 0.1
dirichlet_right = geometry.in_planes(x_nodes, pts[1], [1.,0.])
dirichlet_dofs_right = utility.dof_select(dirichlet_right, jnp.asarray([False, True])) # Only impose y-direction
x_right = x_nodes[dirichlet_right]
v_right = jax.vmap(v_bar)(x_right)
dirichlet_conditions = utility.mask_set(dirichlet_conditions, dirichlet_dofs_right, v_right)

# All dirichlet dofs
dirichlet_dofs = dirichlet_dofs_left + dirichlet_dofs_right

# Generate domain integration points in mesh
(x_int, w_int, n_int, domain_connectivity) = seeder.int_pts_in_tri_mesh(x_nodes, elements, order=1)
max_neighbors_1 = domain_connectivity.shape[-1]

# Definition of pde
lam = 100.
mu = 40.
Em = mu * (3 * lam + 2 * mu) / (lam + mu)
nu = lam / (2 * (lam + mu))
youngs_mod_fun = lambda x: Em
poisson_ratio_fun = lambda x: nu
# pde_1 = models.linear_elasticity_weak(youngs_mod_fun, poisson_ratio_fun, 'plain strain')
pde_1 = models.hyperelastic_steady_state_weak(models.neo_hooke, youngs_mod_fun, poisson_ratio_fun, 'plain strain')


### Setting
n_fields = 2
dofs0 = jnp.zeros((n_nodes, n_fields))
static_settings = flax.core.FrozenDict({
  'solution space': ('fem simplex',),
  'shape function mode': 'direct',
  'number of fields': (n_fields,),
  'assembling mode': ('sparse',),
  'maximal number of neighbors': (max_neighbors_1,),
  'variational scheme': ('weak form galerkin',),
  'model': (pde_1,),
  'solution structure': ('nodal imposition',),
  'solver type': 'newton',
  'solver backend': 'pardiso',
  'solver': 'lu',
  'verbose': 0,
  'dirichlet dofs': utility.jnp_to_tuple(dirichlet_dofs),
  'connectivity': (utility.jnp_to_tuple(domain_connectivity),),
})

settings = {
  'node coordinates': x_nodes,
  'dirichlet conditions': dirichlet_conditions,
  'integration coordinates': (x_int,),
  'integration weights': (w_int,),
}

# Precompute shape functions
# settings = solution_structures.precompile(dofs0, settings, static_settings)


# Call solver
dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8)

# Checks: reaction forces, Dirichlet conditions
print("norm of reaction force: ", jnp.linalg.norm(assembler.assemble_residual(dofs, settings, static_settings)[dirichlet_dofs]))
print("norm of residual at free nodes: ", jnp.linalg.norm(assembler.assemble_residual(dofs, settings, static_settings)[np.invert(dirichlet_dofs)]))
print("dofs at dirichlet nodes: ", dofs[dirichlet_dofs])


### Postprocessing
# Interactive visualization on integration points
local_dofs = dofs[domain_connectivity]

def post_fun(x, itt, local_dofs, settings, static_settings):
  return solution_structures.solution_structure(x, itt, local_dofs, settings, static_settings, 0)

post_fun_vj = jax.jit(jax.vmap(post_fun, (0,0,0,None,None)), static_argnames=['static_settings'])
data = post_fun_vj(x_int, jnp.arange(n_int), local_dofs, settings, static_settings)

x_vis = x_int + jnp.transpose(jnp.asarray([data[:,0],data[:,1]]))

# Plotting with pyvista
plotter.pv_plot(x_vis, data[:,0], export_vtk=False, show=True, data_on_z=False)
plotter.pv_plot(x_vis, data[:,1], export_vtk=False, show=True, data_on_z=False)


# Paraview postprocessing
points = mesh.points
cells = mesh.cells_dict["triangle"]
mesh = meshio.Mesh(
    points,
    {'triangle': cells},
    point_data={
        "u": dofs
    },
)
mesh.write("cooks_membrane.vtk")
